#ifndef MATRIX_HPP
#define MATRIX_HPP

#include "Common.hpp"
#include "Coord.hpp"

template<typename T>
class Matrix
{
public:
    Matrix() : m_res(0), m_data(nullptr) {}
    Matrix(uint32_t res) : m_res(0), m_data(nullptr) { resize(res); }
    Matrix(const Matrix& other) : m_res(other.m_res), m_data(nullptr)
    {
        if (m_res > 0)
        {
            uint32_t size = m_res * m_res * m_res;
            m_data.reset(new T[size]);
            for (size_t i = 0; i < size; i++)
            {
                m_data[i] = other.m_data[i];
            }
        }
    }
    Matrix(Matrix&& other) : m_res(other.m_res), m_data(std::move(other.m_data))
    {
        other.m_res = 0;
    }
    ~Matrix() {}

    void resize(uint32_t res)
    {
        m_res = res;
        if (res > 0)
        {
            uint32_t size = res * res * res;
            m_data.reset(new T[size]());
        }
        else
        {
            m_data.reset();
        }
    }

    uint32_t getRes() const { return m_res; }

    const T& operator()(uint32_t x, uint32_t y, uint32_t z) const
    {
        uint32_t pos = (x * m_res + y) * m_res + z;
        return m_data[pos];
    }
    T& operator()(uint32_t x, uint32_t y, uint32_t z)
    {
        uint32_t pos = (x * m_res + y) * m_res + z;
        return m_data[pos];
    }

    const T& operator()(const Coord& coord) const
    {
        return operator()(coord.x, coord.y, coord.z);
    }
    T& operator()(const Coord& coord)
    {
        return operator()(coord.x, coord.y, coord.z);
    }

private:
    uint32_t m_res;
    std::unique_ptr<T[]> m_data;
};

#endif
