#include "Solve.hpp"
#include "Model.hpp"
#include "Command.hpp"
#include "Matrix.hpp"

using std::vector;
using std::pair;

namespace
{
    /*
    class Column
    {
    public:
        uint8_t m_y;
        uint8_t m_z;
    };

    void findColumns(const Model& model,
                     vector<Column>* pColumns)
    {
        auto& columns = *pColumns;
        uint32_t res = model.getRes();
        uint32_t startY = ((res - 1) % 3 == 1) ? 0 : 1;
        uint32_t startZ = ((res - 2) % 3 == 1) ? 1 : 2;
        for (uint32_t y = startY; y < res - 1; y += 3)
        {
            for (uint32_t z = startZ; z < res - 1; z += 3)
            {
                bool found = false;
                for (uint32_t x = 1; x < res - 1; x++)
                {
                    if (model.get(x, y    , z - 1) ||
                        model.get(x, y    , z    ) ||
                        model.get(x, y    , z + 1) ||
                        model.get(x, y + 1, z - 1) ||
                        model.get(x, y + 1, z    ) ||
                        model.get(x, y + 1, z + 1) ||
                        (y > 0 &&
                         (model.get(x, y - 1, z - 1) ||
                          model.get(x, y - 1, z    ) ||
                          model.get(x, y - 1, z + 1))))
                    {
                        found = true;
                        break;
                    }
                }
                if (found)
                {
                    columns.emplace_back();
                    auto& column = columns.back();
                    column.m_y = y;
                    column.m_z = z;
                }
            }
        }
    }
    */

    class DfsEntry
    {
    public:
        uint8_t m_visited;
        uint8_t m_articulation;
        uint8_t m_important;
        Coord m_parent;
        uint32_t m_depth;
        uint32_t m_low;
    };

    class QueueEntry
    {
    public:
        Coord m_coord;
        uint32_t m_prev;
        Command m_command;
    };

    constexpr Coord adjDeltas[6] =
    {
        Coord(-1, 0, 0),
        Coord(1, 0, 0),
        Coord(0, -1, 0),
        Coord(0, 1, 0),
        Coord(0, 0, -1),
        Coord(0, 0, 1)
    };

    constexpr Coord nearDeltas[18] =
    {
        Coord(-1, -1, 0),
        Coord(0, -1, -1),
        Coord(0, -1, 0),
        Coord(0, -1, 1),
        Coord(1, -1, 0),

        Coord(-1, 0, -1),
        Coord(-1, 0, 0),
        Coord(-1, 0, 1),
        Coord(0, 0, -1),
        Coord(0, 0, 1),
        Coord(1, 0, -1),
        Coord(1, 0, 0),
        Coord(1, 0, 1),

        Coord(-1, 1, 0),
        Coord(0, 1, -1),
        Coord(0, 1, 0),
        Coord(0, 1, 1),
        Coord(1, 1, 0)
    };
}

void solve(const Model& model,
           vector<Command>* pTrace)
{
    /*
    vector<Column> columns;
    findColumns(model, &columns);
    printf("%" PRIuZ " columns\n", columns.size());
    */

    uint32_t res = model.getRes();
    Model scopeModel(res);
    for (uint32_t x = 0; x < res; x++)
    {
        for (uint32_t z = 0; z < res; z++)
        {
            scopeModel.set(x, 0, z, 1);
        }
    }

    //printf("res = %" PRIu32 "\n", res);
    Model readyModel(res);
    uint32_t readyCount = 0;
    for (uint32_t x = 1; x < res - 1; x++)
    {
        for (uint32_t z = 1; z < res - 1; z++)
        {
            if (model.get(x, 0, z))
            {
                readyModel.set(x, 0, z, 1);
                readyCount++;
            }
        }
    }
    //printf("readyCount = %" PRIu32 "\n", readyCount);

    Model buildModel(res);
    Coord coord = Coord();
    auto& trace = *pTrace;
    vector<Coord> dfsStack;
    Matrix<DfsEntry> dfsEntries(res);
    vector<QueueEntry> q; q.reserve(res * res * res);
    Model searchModel(res);
    vector<Command> tmpCommands;

    while (readyCount > 0 || coord != Coord(0, 0, 0))
    {
        // Find articulation points
#if 1
#if 0
        {
            buildModel.set(1, 0, 1, 1);
            buildModel.set(1, 0, 2, 1);
            buildModel.set(1, 0, 3, 1);
            buildModel.set(2, 0, 1, 1);
            readyModel.set(2, 0, 2, 1); // ready
            buildModel.set(2, 0, 3, 1);
            buildModel.set(3, 0, 1, 1);
            buildModel.set(3, 0, 2, 1);
            buildModel.set(3, 0, 3, 1);
            buildModel.set(1, 1, 1, 1);
            buildModel.set(1, 1, 2, 1);
            buildModel.set(1, 1, 3, 1);
            buildModel.set(2, 1, 1, 1);
            readyModel.set(2, 1, 2, 1); // ready
            buildModel.set(2, 1, 3, 1);
            buildModel.set(3, 1, 1, 1);
            buildModel.set(3, 1, 2, 1);
            buildModel.set(3, 1, 3, 1);
        }
#endif
        {
            for (uint32_t x = 0; x < res; x++)
            {
                for (uint32_t y = 0; y < res; y++)
                {
                    for (uint32_t z = 0; z < res; z++)
                    {
                        auto& entry = dfsEntries(x, y, z);
                        entry.m_visited = 0;
                    }
                }
            }

            {
                dfsStack.push_back(Coord(0, 0, 0));
                auto& entry = dfsEntries(Coord(0, 0, 0));
                entry.m_articulation = 0;
                entry.m_important = 0;
                entry.m_parent = Coord(0, 0, 0);
                entry.m_depth = 1;
                entry.m_low = 1;
            }

            uint32_t dfsStartCount = 0;
            uint32_t dfsFinishCount = 0;
            while (!dfsStack.empty())
            {
                Coord dfsCoord = dfsStack.back();
                auto& entry = dfsEntries(dfsCoord);
                if (entry.m_visited == 1)
                {
                    dfsFinishCount++;
                    //printf("< %s\n", dfsCoord.toString().c_str());
                    entry.m_visited = 2;
                    if (entry.m_parent != Coord(0, 0, 0))
                    {
                        auto& parentEntry = dfsEntries(entry.m_parent);
                        //printf("parentEntry.m_low = %" PRIu32 "\n", parentEntry.m_low);
                        //printf("entry.m_depth = %" PRIu32 "\n", entry.m_depth);
                        if (entry.m_low >= parentEntry.m_depth &&
                            entry.m_important)
                        {
                            /*
                            if (!parentEntry.m_articulation)
                            {
                                printf("%s: articulation = 1\n",
                                       entry.m_parent.toString().c_str());
                            }
                            */
                            parentEntry.m_articulation = 1;
                        }
                        parentEntry.m_low = std::min(parentEntry.m_low, entry.m_low);
                        if (entry.m_important)
                        {
                            parentEntry.m_important = 1;
                        }
                    }
                }
                if (entry.m_visited)
                {
                    dfsStack.pop_back();
                    continue;
                }
                dfsStartCount++;
                //printf("> %s\n", dfsCoord.toString().c_str());
                entry.m_visited = 1;
                entry.m_articulation = 0;

                for (uint32_t adjDir = 0; adjDir < 6; adjDir++)
                {
                    Coord adjCoord = dfsCoord + adjDeltas[adjDir];
                    if (adjCoord.x >= res ||
                        adjCoord.y >= res ||
                        adjCoord.z >= res)
                    {
                        continue;
                    }
                    if (!scopeModel.get(adjCoord) ||
                        buildModel.get(adjCoord))
                    {
                        continue;
                    }
                    auto& adjEntry = dfsEntries(adjCoord);
                    if (!adjEntry.m_visited)
                    {
                        adjEntry.m_parent = dfsCoord;
                        adjEntry.m_depth = entry.m_depth + 1;
                        adjEntry.m_low = entry.m_depth + 1;
                        adjEntry.m_important =
                            model.get(adjCoord) ||
                            adjCoord == coord;
                        dfsStack.push_back(adjCoord);
                    }
                    else if (adjCoord != entry.m_parent)
                    {
                        entry.m_low = std::min(entry.m_low, adjEntry.m_depth);
                    }
                }
            }
            //printf("dfsStartCount = %" PRIu32 "\n", dfsStartCount);
            //printf("dfsFinishCount = %" PRIu32 "\n", dfsFinishCount);

            /*
            for (uint32_t x = 0; x < res; x++)
            {
                for (uint32_t y = 0; y < res; y++)
                {
                    for (uint32_t z = 0; z < res; z++)
                    {
                        if (model.get(x, y, z) &&
                            !buildModel.get(x, y, z) &&
                            !dfsEntries(x, y, z).m_visited)
                        {
                            printf("Didn't visit %s\n",
                                   Coord(x, y, z).toString().c_str());
                        }
                    }
                }
            }
            */
        }
#if 0
        exit(0);
#endif
#endif

        for (auto& entry : q)
        {
            searchModel.set(entry.m_coord.x,
                            entry.m_coord.y,
                            entry.m_coord.z,
                            0);
        }
        q.clear();

        size_t qpos = 0;

        {
            q.emplace_back();
            auto& entry = q.back();
            entry.m_coord = coord;
            entry.m_prev = 0;
            entry.m_command = Command(Action::Halt);
            searchModel.set(coord.x, coord.y, coord.z, 1);
        }

        Coord readyCoord;
        Coord readyDelta;
        size_t opos = 0;
        bool foundOrigin = false;
        while (qpos < q.size())
        {
            uint32_t x = q[qpos].m_coord.x;
            uint32_t y = q[qpos].m_coord.y;
            uint32_t z = q[qpos].m_coord.z;
            //printf("x = %" PRIu32 ", y = %" PRIu32 ", z = %" PRIu32 "\n", x, y, z);

            if (q[qpos].m_coord == Coord(0, 0, 0))
            {
                opos = qpos;
                foundOrigin = true;
            }

            if (readyCount == 0)
            {
                if (q[qpos].m_coord == Coord(0, 0, 0))
                {
                    break;
                }
            }
            else
            {
                bool foundReady = false;
                for (uint32_t nearDir = 0; nearDir < 18; nearDir++)
                {
                    Coord nearCoord = q[qpos].m_coord + nearDeltas[nearDir];
                    if (nearCoord.x >= res ||
                        nearCoord.y >= res ||
                        nearCoord.z >= res)
                    {
                        continue;
                    }
                    if (readyModel.get(nearCoord) &&
                        !dfsEntries(nearCoord).m_articulation)
                    {
                        foundReady = true;
                        readyCoord = nearCoord;
                        readyDelta = nearDeltas[nearDir];
                        break;
                    }
                }

                if (foundReady)
                {
                    break;
                }
            }

            for (uint32_t dist = 1; dist <= 15; dist++)
            {
                if (x < dist || buildModel.get(x - dist, y, z))
                {
                    break;
                }
                if (!searchModel.get(x - dist, y, z))
                {
                    q.emplace_back();
                    auto& entry = q.back();
                    entry.m_coord = Coord(x - dist, y, z);
                    //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                    entry.m_prev = qpos;
                    entry.m_command = Command(Action::SMove,
                                              Coord(-dist, 0, 0));
                    searchModel.set(x - dist, y, z, 1);
                }
            }
            for (uint32_t dist = 1; dist <= 15; dist++)
            {
                if (x + dist >= res || buildModel.get(x + dist, y, z))
                {
                    break;
                }
                if (!searchModel.get(x + dist, y, z))
                {
                    q.emplace_back();
                    auto& entry = q.back();
                    entry.m_coord = Coord(x + dist, y, z);
                    //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                    entry.m_prev = qpos;
                    entry.m_command = Command(Action::SMove,
                                              Coord(dist, 0, 0));
                    searchModel.set(x + dist, y, z, 1);
                }
            }
            for (uint32_t dist = 1; dist <= 15; dist++)
            {
                if (y < dist || buildModel.get(x, y - dist, z))
                {
                    break;
                }
                if (!searchModel.get(x, y - dist, z))
                {
                    q.emplace_back();
                    auto& entry = q.back();
                    entry.m_coord = Coord(x, y - dist, z);
                    //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                    entry.m_prev = qpos;
                    entry.m_command = Command(Action::SMove,
                                              Coord(0, -dist, 0));
                    searchModel.set(x, y - dist, z, 1);
                }
            }
            for (uint32_t dist = 1; dist <= 15; dist++)
            {
                if (y + dist >= res || buildModel.get(x, y + dist, z))
                {
                    break;
                }
                if (!searchModel.get(x, y + dist, z))
                {
                    q.emplace_back();
                    auto& entry = q.back();
                    entry.m_coord = Coord(x, y + dist, z);
                    //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                    entry.m_prev = qpos;
                    entry.m_command = Command(Action::SMove,
                                              Coord(0, dist, 0));
                    searchModel.set(x, y + dist, z, 1);
                }
            }
            for (uint32_t dist = 1; dist <= 15; dist++)
            {
                if (z < dist || buildModel.get(x, y, z - dist))
                {
                    break;
                }
                if (!searchModel.get(x, y, z - dist))
                {
                    q.emplace_back();
                    auto& entry = q.back();
                    entry.m_coord = Coord(x, y, z - dist);
                    //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                    entry.m_prev = qpos;
                    entry.m_command = Command(Action::SMove,
                                              Coord(0, 0, -dist));
                    searchModel.set(x, y, z - dist, 1);
                }
            }
            for (uint32_t dist = 1; dist <= 15; dist++)
            {
                if (z + dist >= res || buildModel.get(x, y, z + dist))
                {
                    break;
                }
                if (!searchModel.get(x, y, z + dist))
                {
                    q.emplace_back();
                    auto& entry = q.back();
                    entry.m_coord = Coord(x, y, z + dist);
                    //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                    entry.m_prev = qpos;
                    entry.m_command = Command(Action::SMove,
                                              Coord(0, 0, dist));
                    searchModel.set(x, y, z + dist, 1);
                }
            }

            qpos++;
        }

        if (qpos == q.size() && coord != Coord(0, 0, 0))
        {
            if (!foundOrigin)
            {
                printf("!foundOrigin\n");
                break;
            }
            while (q[opos].m_prev != 0)
            {
                opos = q[opos].m_prev;
            }
            auto& entry = q[opos];
            trace.push_back(entry.m_command);
            printf("Coord %s -> %s\n",
                   coord.toString().c_str(),
                   q[opos].m_coord.toString().c_str());
            coord = q[opos].m_coord;
            continue;
        }

        if (qpos == q.size())
        {
            break;
        }

        coord = q[qpos].m_coord;
        //printf("coord = %s\n", coord.toString().c_str());

        while (qpos != 0)
        {
            auto& entry = q[qpos];
            tmpCommands.push_back(entry.m_command);
            qpos = entry.m_prev;
        }
        /*
        for (auto it = tmpCommands.rbegin(); it != tmpCommands.rend(); ++it)
        {
            printf("%s\n", it->toString().c_str());
        }
        */
        trace.insert(trace.end(), tmpCommands.rbegin(), tmpCommands.rend());
        tmpCommands.clear();

        if (readyCount == 0)
        {
            continue;
        }

        //printf("readyCoord = %s\n", readyCoord.toString().c_str());
        readyModel.set(readyCoord, 0);
        buildModel.set(readyCoord, 1);
        readyCount--;
        for (uint32_t adjDir = 0; adjDir < 6; adjDir++)
        {
            Coord adjCoord = readyCoord + adjDeltas[adjDir];
            if (adjCoord.x >= res ||
                adjCoord.y >= res ||
                adjCoord.z >= res)
            {
                continue;
            }
            if (model.get(adjCoord) &&
                !buildModel.get(adjCoord) &&
                !readyModel.get(adjCoord))
            {
                readyModel.set(adjCoord, 1);
                readyCount++;
            }
        }

        trace.push_back(Command(Action::Fill, readyDelta));
        //printf("%s\n", trace.back().toString().c_str());

        for (uint32_t x = 0; x < 3; x++)
        {
            for (uint32_t y = 0; y < 3; y++)
            {
                for (uint32_t z = 0; z < 3; z++)
                {
                    Coord scopeCoord = readyCoord + Coord(x - 1, y - 1, z - 1);
                    if (scopeCoord.x < res &&
                        scopeCoord.y < res &&
                        scopeCoord.z < res)
                    {
                        scopeModel.set(scopeCoord, 1);
                    }
                }
            }
        }
    }

    if (readyCount > 0)
    {
        printf("readyCount = %" PRIu32 "\n", readyCount);
        printf("Painted into a corner\n");
        trace.clear();
        return;
    }

    trace.push_back(Command(Action::Halt));
}
