#include "Check.hpp"
#include "Model.hpp"
#include "Command.hpp"
#include "Trace.hpp"
#include "Sim.hpp"

using std::string;
using std::vector;

bool checkTrace(bool full,
                const Model& sourceModel,
                const Model& targetModel,
                const vector<Command>& trace,
                uint64_t* pEnergy,
                string* pMsg)
{
    {
        vector<uint8_t> data;
        if (!formatTrace(full, trace, &data, pMsg))
        {
            return false;
        }
    }

    Sim sim(full);
    sim.init(sourceModel);
    size_t pos = 0;
    while (pos < trace.size())
    {
        if (sim.m_bots.empty())
        {
            if (pMsg) *pMsg = "Halted before end of trace";
            return false;
        }
        if (!sim.step(trace, pos, pMsg))
        {
            return false;
        }
    }

    if (!sim.m_bots.empty())
    {
        if (pMsg) *pMsg = "Trace did not halt";
        return false;
    }
    if (sim.m_model != targetModel)
    {
        if (pMsg) *pMsg = "Models not equal";
        return false;
    }

    if (pEnergy) *pEnergy = sim.m_energy;

    return true;
}
