#ifndef COORD_HPP
#define COORD_HPP

#include "Common.hpp"

class Coord
{
public:
    constexpr Coord() :
        x(0), y(0), z(0) {}
    constexpr Coord(uint8_t x_, uint8_t y_, uint8_t z_) :
        x(x_), y(y_), z(z_) {}

    Coord& operator+=(const Coord& other)
    {
        x += other.x;
        y += other.y;
        z += other.z;
        return *this;
    }

    std::string toString() const;

    uint8_t x;
    uint8_t y;
    uint8_t z;
};

inline Coord operator+(const Coord& a, const Coord& b)
{
    return Coord(a.x + b.x, a.y + b.y, a.z + b.z);
}

inline Coord operator-(const Coord& a, const Coord& b)
{
    return Coord(a.x - b.x, a.y - b.y, a.z - b.z);
}

inline bool operator==(const Coord& a, const Coord& b)
{
    return a.x == b.x && a.y == b.y && a.z == b.z;
}

inline bool operator!=(const Coord& a, const Coord& b)
{
    return a.x != b.x || a.y != b.y || a.z != b.z;
}

inline bool operator<(const Coord& a, const Coord& b)
{
    if (a.x != b.x) return a.x < b.x;
    if (a.y != b.y) return a.y < b.y;
    return a.z < b.z;
}

#endif
