#include "Mapping.hpp"
#include "StringUtils.hpp"
#include "Model.hpp"

using std::string;
using std::vector;

vector<Mapping> getMappings(Batch batch)
{
    string msg;

    if (batch != Batch::FRD &&
        batch != Batch::FRA)
    {
        printf("getMappings: invalid batch\n");
        exit(1);
    }

    vector<Mapping> result;

    Batch fadBatch = (batch == Batch::FRD) ? Batch::FD : Batch::FA;
    Model fadModels[186];

    for (uint32_t i = 1; i <= 186; i++)
    {
        string fadModelName =
            fadBatch == Batch::FD ?
            strprintf("../../../data/problemsF/FD%03" PRIu32 "_src.mdl", i) :
            strprintf("../../../data/problemsF/FA%03" PRIu32 "_tgt.mdl", i);
        if (!fadModels[i - 1].load(fadModelName, &msg))
        {
            printf("%03" PRIu32": %s\n", i, msg.c_str());
            exit(1);
        }
    }

    Model frModels[115];

    for (uint32_t i = 1; i <= 115; i++)
    {
        string frModelName =
            batch == Batch::FRD ?
            strprintf("../../../data/problemsF/FR%03" PRIu32 "_src.mdl", i) :
            strprintf("../../../data/problemsF/FR%03" PRIu32 "_tgt.mdl", i);
        if (!frModels[i - 1].load(frModelName, &msg))
        {
            printf("%03" PRIu32": %s\n", i, msg.c_str());
            exit(1);
        }

        bool found = false;

        {
            for (uint32_t j = 1; j <= 186; j++)
            {
                if (frModels[i - 1] == fadModels[j - 1])
                {
                    found = true;
                    result.push_back(Mapping(fadBatch, j));
                    break;
                }
            }
        }

        if (!found)
        {
            for (uint32_t j = 1; j <= i; j++)
            {
                if (frModels[i - 1] == frModels[j - 1])
                {
                    found = true;
                    result.push_back(Mapping(batch, j));
                    break;
                }
            }
        }

        if (!found)
        {
            printf("!found\n");
            exit(1);
        }
    }

    return result;
}
