#ifndef MODEL_HPP
#define MODEL_HPP

#include "Common.hpp"
#include "Coord.hpp"

class Model
{
public:
    Model() : m_res(0), m_data(nullptr) {}
    Model(uint32_t res) : m_res(0), m_data(nullptr) { resize(res); }
    Model(const Model& other);
    Model(Model&& other);
    ~Model() {}

    Model& operator=(const Model& other);
    Model& operator=(Model&& other);

    void resize(uint32_t res);

    uint32_t getRes() const { return m_res; }

    uint8_t get(uint32_t x, uint32_t y, uint32_t z) const
    {
        uint32_t pos = (x * m_res + y) * m_res + z;
        uint32_t bytePos = pos >> 3;
        uint32_t bitPos = pos & 7;
        return (m_data[bytePos] >> bitPos) & (uint8_t)1;
    }
    void set(uint32_t x, uint32_t y, uint32_t z, uint8_t value)
    {
        uint32_t pos = (x * m_res + y) * m_res + z;
        uint32_t bytePos = pos >> 3;
        uint32_t bitPos = pos & 7;
        if (value)
        {
            m_data[bytePos] |= ((uint8_t)1 << bitPos);
        }
        else
        {
            m_data[bytePos] &= ~((uint8_t)1 << bitPos);
        }
    }

    uint8_t get(const Coord& coord) const
    {
        return get(coord.x, coord.y, coord.z);
    }
    void set(const Coord& coord, uint8_t value)
    {
        set(coord.x, coord.y, coord.z, value);
    }

    bool load(const std::string& fileName,
              std::string* pMsg = nullptr);

    bool empty() const;

    bool operator==(const Model& other) const;
    bool operator!=(const Model& other) const { return !(*this == other); }

private:
    uint32_t m_res;
    std::unique_ptr<uint8_t[]> m_data;
};

#endif
