#ifndef SIM_HPP
#define SIM_HPP

#include "Common.hpp"
#include "Coord.hpp"
#include "Model.hpp"

class Command;
class FusionRec;
class GroupRec;

class Bot
{
public:
    uint8_t m_bid = 0;
    Coord m_pos;
    std::vector<uint8_t> m_seeds;
};

class Sim
{
public:
    Sim(bool full);
    Sim(const Sim&) = delete;
    Sim(Sim&&) = delete;
    ~Sim();
    Sim& operator=(const Sim&) = delete;
    Sim& operator=(Sim&&) = delete;

    void init(uint32_t res);
    void init(const Model& model);
    void init(Model&& model);

    bool step(const std::vector<Command>& trace,
              size_t& pos,
              std::string* pMsg = nullptr);

private:
    void initImpl();

public:
    bool m_full;
    uint64_t m_energy;
    bool m_harmonics;
    Model m_model;
    std::vector<Bot> m_bots;

private:
    std::vector<Coord> volatileCoords;
    std::vector<Bot> newBots;
    std::vector<std::pair<Coord, Coord>> fills;
    std::vector<std::pair<Coord, Coord>> voids;
    std::vector<FusionRec> fusionRecs;
    std::vector<GroupRec> groupRecs;
};

#endif
