#include "Solve.hpp"
#include "Model.hpp"
#include "Command.hpp"
#include "Matrix.hpp"
#include <random>

using std::vector;
using std::pair;

#define USE_SCOPE_MODEL 0
#define USE_SAFE_COUNT 1
#define SAFE_COUNT_FRACTION 0
#define MAX_BOTS_FACTOR 2
#define USE_MULTI 1
#define USE_40 1
#define USE_HARMONICS 1
#define MAX_GROUPS 0
#define ERECT 1
#define USE_TASKS 1

namespace
{
    void findBBox(const Model& model,
                  Coord* pCoordLow,
                  Coord* pCoordHigh)
    {
        uint32_t res = model.getRes();
        Coord coordLow(res - 1, res - 1, res - 1);
        Coord coordHigh(0, 0, 0);

        for (uint32_t x = 0; x < res; x++)
        {
            for (uint32_t y = 0; y < res; y++)
            {
                for (uint32_t z = 0; z < res; z++)
                {
                    if (model.get(x, y, z))
                    {
                        if (x < coordLow.x) coordLow.x = x;
                        if (y < coordLow.y) coordLow.y = y;
                        if (z < coordLow.z) coordLow.z = z;
                        if (x > coordHigh.x) coordHigh.x = x;
                        if (y > coordHigh.y) coordHigh.y = y;
                        if (z > coordHigh.z) coordHigh.z = z;
                    }
                }
            }
        }

        *pCoordLow = coordLow;
        *pCoordHigh = coordHigh;
    }

    class Column
    {
    public:
        Coord m_coordLow;
        Coord m_coordHigh;
    };

    void findColumnsD(const Model& model,
                      vector<Column>* pColumns)
    {
        auto& columns = *pColumns;
        uint32_t res = model.getRes();

        uint32_t startZ = 1;
        while (true)
        {
            // Find startZ

            while (startZ < res - 1)
            {
                bool foundVoxel = false;
                for (uint32_t x = 1; x < res - 1; x++)
                {
                    for (uint32_t y = 0; y < res - 1; y++)
                    {
                        if (model.get(x, y, startZ))
                        {
                            foundVoxel = true;
                            break;
                        }
                    }
                    if (foundVoxel)
                    {
                        break;
                    }
                }
                if (foundVoxel)
                {
                    break;
                }
                startZ++;
            }
            if (startZ >= res - 1)
            {
                break;
            }

            // Find endZ

            uint32_t endZ = startZ + 30;
            if (endZ > res - 2)
            {
                endZ = res - 2;
            }
            while (true)
            {
                bool foundVoxel = false;
                for (uint32_t x = 1; x < res - 1; x++)
                {
                    for (uint32_t y = 0; y < res - 1; y++)
                    {
                        if (model.get(x, y, endZ))
                        {
                            foundVoxel = true;
                            break;
                        }
                    }
                    if (foundVoxel)
                    {
                        break;
                    }
                }
                if (foundVoxel)
                {
                    break;
                }
                endZ--;
            }
            if (endZ == startZ)
            {
                endZ++;
            }

            uint32_t startX = 1;
            while (true)
            {
                // Find startX

                while (startX < res - 1)
                {
                    bool foundVoxel = false;
                    for (uint32_t y = 0; y < res - 1; y++)
                    {
                        for (uint32_t z = startZ; z <= endZ; z++)
                        {
                            if (model.get(startX, y, z))
                            {
                                foundVoxel = true;
                                break;
                            }
                        }
                        if (foundVoxel)
                        {
                            break;
                        }
                    }
                    if (foundVoxel)
                    {
                        break;
                    }
                    startX++;
                }
                if (startX >= res - 1)
                {
                    break;
                }

                // Find endX

                uint32_t endX = startX + 30;
                if (endX > res - 2)
                {
                    endX = res - 2;
                }
                while (true)
                {
                    bool foundVoxel = false;
                    for (uint32_t y = 0; y < res - 1; y++)
                    {
                        for (uint32_t z = startZ; z <= endZ; z++)
                        {
                            if (model.get(endX, y, z))
                            {
                                foundVoxel = true;
                                break;
                            }
                        }
                        if (foundVoxel)
                        {
                            break;
                        }
                    }
                    if (foundVoxel)
                    {
                        break;
                    }
                    endX--;
                }
                if (endX == startX)
                {
                    endX++;
                }

                // Find startY

                uint32_t startY = 0;
                while (true)
                {
                    bool foundVoxel = false;
                    for (uint32_t x = startX; x <= endX; x++)
                    {
                        for (uint32_t z = startZ; z <= endZ; z++)
                        {
                            if (model.get(x, startY, z))
                            {
                                foundVoxel = true;
                                break;
                            }
                        }
                        if (foundVoxel)
                        {
                            break;
                        }
                    }
                    if (foundVoxel)
                    {
                        break;
                    }
                    startY++;
                }

                // Find endY

                uint32_t endY = res - 1;
                while (true)
                {
                    bool foundVoxel = false;
                    for (uint32_t x = startX; x <= endX; x++)
                    {
                        for (uint32_t z = startZ; z <= endZ; z++)
                        {
                            if (model.get(x, endY, z))
                            {
                                foundVoxel = true;
                                break;
                            }
                        }
                        if (foundVoxel)
                        {
                            break;
                        }
                    }
                    if (foundVoxel)
                    {
                        break;
                    }
                    endY--;
                }

                // Generate column

                {
                    columns.emplace_back();
                    auto& column = columns.back();
                    column.m_coordLow = Coord(startX, startY, startZ);
                    column.m_coordHigh = Coord(endX, endY, endZ);
                }

                startX = endX + 1;
            }

            startZ = endZ + 1;
        }

        std::stable_sort(columns.begin(), columns.end(),
                         [](const Column& a, const Column& b)
                         {
                             uint32_t heightA = a.m_coordHigh.y - a.m_coordLow.y;
                             uint32_t heightB = b.m_coordHigh.y - b.m_coordLow.y;
                             return heightA > heightB;
                         });
    }

    enum class GroupState : uint32_t
    {
        Move, Void, Home
    };

    class Task
    {
    public:
        Coord m_coordLow;
        Coord m_coordHigh;
    };

    void findTasks(const Model& model,
                   vector<uint8_t>& scratch,
                   uint32_t level,
                   vector<Task>* pTasks,
                   uint32_t* pNumTasked,
                   uint32_t* pNumUntasked)
    {
        auto& tasks = *pTasks;
        uint32_t res = model.getRes();
        scratch.clear();
        scratch.resize(res * res);
        uint32_t numTasked = 0;
        uint32_t numUntasked = 0;
        for (uint32_t lowX = 1; lowX < res - 1; lowX++)
        {
            for (uint32_t lowZ = 1; lowZ < res - 1; lowZ++)
            {
                if (!model.get(lowX, level, lowZ) ||
                    scratch[lowX * res + lowZ])
                {
                    continue;
                }
                numUntasked++;
                uint32_t highX = lowX;
                uint32_t highZ = lowZ;
                bool doneX = false;
                bool doneZ = false;
                while (!doneX || !doneZ)
                {
                    if (!doneX)
                    {
                        if (highX == lowX + 30)
                        {
                            doneX = true;
                        }
                    }
                    if (!doneX)
                    {
                        for (uint32_t z = lowZ; z <= highZ; z++)
                        {
                            if (!model.get(highX + 1, level, z) ||
                                scratch[(highX + 1) * res + z])
                            {
                                doneX = true;
                                break;
                            }
                        }
                    }
                    if (!doneX)
                    {
                        highX++;
                    }

                    if (!doneZ)
                    {
                        if (highZ == lowZ + 30)
                        {
                            doneZ = true;
                        }
                    }
                    if (!doneZ)
                    {
                        for (uint32_t x = lowX; x <= highX; x++)
                        {
                            if (!model.get(x, level, highZ + 1) ||
                                scratch[x * res + (highZ + 1)])
                            {
                                doneZ = true;
                            }
                        }
                    }
                    if (!doneZ)
                    {
                        highZ++;
                    }
                }
                if (highX == lowX || highZ == lowZ)
                {
                    continue;
                }
                if ((highX - lowX) * (highZ - lowZ) < 15)
                {
                    continue;
                }

                numUntasked--;
                numTasked += (highX - lowX + 1) * (highZ - lowZ + 1);
                {
                    tasks.emplace_back();
                    auto& task = tasks.back();
                    task.m_coordLow = Coord(lowX, level, lowZ);
                    task.m_coordHigh = Coord(highX, level, highZ);
                }
                for (uint32_t x = lowX; x <= highX; x++)
                {
                    for (uint32_t z = lowZ; z <= highZ; z++)
                    {
                        scratch[x * res + z] = 1;
                    }
                }
            }
        }
        if (pNumTasked) *pNumTasked = numTasked;
        if (pNumUntasked) *pNumUntasked = numUntasked;
    }

    enum class TaskState : uint32_t
    {
        Move, Fill, None
    };

    class DfsEntry
    {
    public:
        uint8_t m_visited;
        uint8_t m_articulation;
        uint8_t m_important;
        Coord m_parent;
        uint32_t m_depth;
        uint32_t m_low;
    };

    class QueueEntry
    {
    public:
        Coord m_coord;
        uint32_t m_prev;
        Command m_command;
    };

    constexpr Coord adjDeltas[6] =
    {
        Coord(-1, 0, 0),
        Coord(1, 0, 0),
        Coord(0, -1, 0),
        Coord(0, 1, 0),
        Coord(0, 0, -1),
        Coord(0, 0, 1)
    };

    constexpr Coord nearDeltas[18] =
    {
        Coord(-1, -1, 0),
        Coord(0, -1, -1),
        Coord(0, -1, 0),
        Coord(0, -1, 1),
        Coord(1, -1, 0),

        Coord(-1, 0, -1),
        Coord(-1, 0, 0),
        Coord(-1, 0, 1),
        Coord(0, 0, -1),
        Coord(0, 0, 1),
        Coord(1, 0, -1),
        Coord(1, 0, 0),
        Coord(1, 0, 1),

        Coord(-1, 1, 0),
        Coord(0, 1, -1),
        Coord(0, 1, 0),
        Coord(0, 1, 1),
        Coord(1, 1, 0)
    };
}

void solve(bool full,
           const Model& sourceModel,
           const Model& targetModel,
           vector<Command>* pTrace)
{
    std::default_random_engine gen(2);

    auto& trace = *pTrace;

    bool disassemble = !sourceModel.empty();
    bool assemble = !targetModel.empty();
    if (disassemble && assemble)
    {
        // Not supported
        return;
    }

    vector<Column> columnsD;
    if (disassemble)
    {
        findColumnsD(sourceModel, &columnsD);
#if 0
        printf("\n");
        printf("%" PRIuZ " columns\n", columnsD.size());
        for (auto& column : columnsD)
        {
            printf("  %s - %s\n",
                   column.m_coordLow.toString().c_str(),
                   column.m_coordHigh.toString().c_str());
        }
        //return;
#endif
    }
    size_t numColumnsD = columnsD.size();

    bool erect = ERECT && assemble;

    bool useMulti = USE_MULTI || disassemble;
    bool use40 = USE_40 ? full : false;
    if (disassemble && numColumnsD <= 5)
    {
        use40 = false;
    }
    bool useHarmonics = USE_HARMONICS && (disassemble || (assemble && erect));
    bool useScopeModel = USE_SCOPE_MODEL && (assemble && !erect);

    if (useMulti)
    {
        // Fission

        //  1   2-40  ( 0,  0,  0)

        if (use40)
        {
            trace.push_back(Command(Action::Fission, Coord(0, 1, 0), 19));
        }

        //  1  22-40  ( 0,  0,  0)
        //  2   3-21  ( 0,  1,  0)

        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 9));
        if (use40)
        {
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 9));
        }

        //  1  32-40  ( 0,  0,  0)
        //  2  13-21  ( 0,  1,  0)
        //  3   4-12  ( 0,  1,  1)
        // 22  23-31  ( 1,  0,  0)

        trace.push_back(Command(Action::Wait));
        if (use40)
        {
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::SMove, Coord(0, 0, 9)));
        }
        trace.push_back(Command(Action::SMove, Coord(9, 0, 0)));

        //  1  32-40  ( 0,  0,  0)
        //  2  13-21  ( 0,  1,  0)
        //  3   4-12  ( 0,  1, 10)
        // 22  23-31  (10,  0,  0)

        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 4));
        if (use40)
        {
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 4));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 4));
        }
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 4));

        //  1  37-40  ( 0,  0,  0)
        //  2  18-21  ( 0,  1,  0)
        //  3   9-12  ( 0,  1, 10)
        //  4   5- 8  ( 0,  1, 11)
        // 13  14-17  ( 0,  1,  1)
        // 22  28-31  (10,  0,  0)
        // 23  24-27  (11,  0,  0)
        // 32  33-36  ( 1,  0,  0)

        trace.push_back(Command(Action::Wait));
        if (use40)
        {
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::SMove, Coord(0, 0, 4)));
            trace.push_back(Command(Action::SMove, Coord(0, 0, 4)));
        }
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::SMove, Coord(4, 0, 0)));
        trace.push_back(Command(Action::SMove, Coord(4, 0, 0)));

        //  1  37-40  ( 0,  0,  0)
        //  2  18-21  ( 0,  1,  0)
        //  3   9-12  ( 0,  1, 10)
        //  4   5- 8  ( 0,  1, 15)
        // 13  14-17  ( 0,  1,  5)
        // 22  28-31  (10,  0,  0)
        // 23  24-27  (15,  0,  0)
        // 32  33-36  ( 5,  0,  0)

        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 2));
        if (use40)
        {
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 2));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 2));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 2));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 2));
        }
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 2));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 2));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 2));

        //  1  40-40  ( 0,  0,  0)
        //  2  21-21  ( 0,  1,  0)
        //  3  12-12  ( 0,  1, 10)
        //  4   8- 8  ( 0,  1, 15)
        //  5   6- 7  ( 0,  1, 16)
        //  9  10-11  ( 0,  1, 11)
        // 13  17-17  ( 0,  1,  5)
        // 14  15-16  ( 0,  1,  6)
        // 18  19-20  ( 0,  1,  1)
        // 22  31-31  (10,  0,  0)
        // 23  27-27  (15,  0,  0)
        // 24  25-26  (16,  0,  0)
        // 28  29-30  (11,  0,  0)
        // 32  36-36  ( 5,  0,  0)
        // 33  34-35  ( 6,  0,  0)
        // 37  38-39  ( 1,  0,  0)

        trace.push_back(Command(Action::Wait));
        if (use40)
        {
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::SMove, Coord(0, 0, 1)));
            trace.push_back(Command(Action::SMove, Coord(0, 0, 1)));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::SMove, Coord(0, 0, 1)));
            trace.push_back(Command(Action::SMove, Coord(0, 0, 1)));
        }
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::SMove, Coord(1, 0, 0)));
        trace.push_back(Command(Action::SMove, Coord(1, 0, 0)));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::SMove, Coord(1, 0, 0)));
        trace.push_back(Command(Action::SMove, Coord(1, 0, 0)));

        //  1  40-40  ( 0,  0,  0)
        //  2  21-21  ( 0,  1,  0)
        //  3  12-12  ( 0,  1, 10)
        //  4   8- 8  ( 0,  1, 15)
        //  5   6- 7  ( 0,  1, 17)
        //  9  10-11  ( 0,  1, 12)
        // 13  17-17  ( 0,  1,  5)
        // 14  15-16  ( 0,  1,  7)
        // 18  19-20  ( 0,  1,  2)
        // 22  31-31  (10,  0,  0)
        // 23  27-27  (15,  0,  0)
        // 24  25-26  (17,  0,  0)
        // 28  29-30  (12,  0,  0)
        // 32  36-36  ( 5,  0,  0)
        // 33  34-35  ( 7,  0,  0)
        // 37  38-39  ( 2,  0,  0)

        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 0));
        if (use40)
        {
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 0));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 0));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 0));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 1));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 1));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 0));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 1));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 1));
        }
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 0));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 0));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 1));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 1));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 0));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 1));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 1));

        //  1         ( 0,  0,  0)
        //  2         ( 0,  1,  0)
        //  3         ( 0,  1, 10)
        //  4         ( 0,  1, 15)
        //  5         ( 0,  1, 17)
        //  6   7- 7  ( 0,  1, 18)
        //  8         ( 0,  1, 16)
        //  9         ( 0,  1, 12)
        // 10  11-11  ( 0,  1, 13)
        // 12         ( 0,  1, 11)
        // 13         ( 0,  1,  5)
        // 14         ( 0,  1,  7)
        // 15  16-16  ( 0,  1,  8)
        // 17         ( 0,  1,  6)
        // 18         ( 0,  1,  2)
        // 19  20-20  ( 0,  1,  3)
        // 21         ( 0,  1,  1)
        // 22         (10,  0,  0)
        // 23         (15,  0,  0)
        // 24         (17,  0,  0)
        // 25  26-26  (18,  0,  0)
        // 27         (16,  0,  0)
        // 28         (12,  0,  0)
        // 29  30-30  (13,  0,  0)
        // 31         (11,  0,  0)
        // 32         ( 5,  0,  0)
        // 33         ( 7,  0,  0)
        // 34  35-35  ( 8,  0,  0)
        // 36         ( 6,  0,  0)
        // 37         ( 2,  0,  0)
        // 38  39-39  ( 3,  0,  0)
        // 40         ( 1,  0,  0)

        trace.push_back(Command(useHarmonics ? Action::Flip : Action::Wait));
        if (use40)
        {
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 0));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 0));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 0));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Fission, Coord(0, 0, 1), 0));
            trace.push_back(Command(Action::Wait));
        }
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 0));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 0));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 0));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Fission, Coord(1, 0, 0), 0));
        trace.push_back(Command(Action::Wait));

        //  1         ( 0,  0,  0)
        //  2         ( 0,  1,  0)
        //  3         ( 0,  1, 10)
        //  4         ( 0,  1, 15)
        //  5         ( 0,  1, 17)
        //  6         ( 0,  1, 18)
        //  7         ( 0,  1, 19)
        //  8         ( 0,  1, 16)
        //  9         ( 0,  1, 12)
        // 10         ( 0,  1, 13)
        // 11         ( 0,  1, 14)
        // 12         ( 0,  1, 11)
        // 13         ( 0,  1,  5)
        // 14         ( 0,  1,  7)
        // 15         ( 0,  1,  8)
        // 16         ( 0,  1,  9)
        // 17         ( 0,  1,  6)
        // 18         ( 0,  1,  2)
        // 19         ( 0,  1,  3)
        // 20         ( 0,  1,  4)
        // 21         ( 0,  1,  1)
        // 22         (10,  0,  0)
        // 23         (15,  0,  0)
        // 24         (17,  0,  0)
        // 25         (18,  0,  0)
        // 26         (19,  0,  0)
        // 27         (16,  0,  0)
        // 28         (12,  0,  0)
        // 29         (13,  0,  0)
        // 30         (14,  0,  0)
        // 31         (11,  0,  0)
        // 32         ( 5,  0,  0)
        // 33         ( 7,  0,  0)
        // 34         ( 8,  0,  0)
        // 35         ( 9,  0,  0)
        // 36         ( 6,  0,  0)
        // 37         ( 2,  0,  0)
        // 38         ( 3,  0,  0)
        // 39         ( 4,  0,  0)
        // 40         ( 1,  0,  0)
    }

    size_t numBots = useMulti ? (use40 ? 40 : 20) : 1;
    vector<Coord> botHomes; botHomes.reserve(numBots);
    if (useMulti)
    {
        botHomes.push_back(Coord( 0,  0,  0));
        if (use40)
        {
            botHomes.push_back(Coord( 0,  1,  0));
            botHomes.push_back(Coord( 0,  1, 10));
            botHomes.push_back(Coord( 0,  1, 15));
            botHomes.push_back(Coord( 0,  1, 17));
            botHomes.push_back(Coord( 0,  1, 18));
            botHomes.push_back(Coord( 0,  1, 19));
            botHomes.push_back(Coord( 0,  1, 16));
            botHomes.push_back(Coord( 0,  1, 12));
            botHomes.push_back(Coord( 0,  1, 13));
            botHomes.push_back(Coord( 0,  1, 14));
            botHomes.push_back(Coord( 0,  1, 11));
            botHomes.push_back(Coord( 0,  1,  5));
            botHomes.push_back(Coord( 0,  1,  7));
            botHomes.push_back(Coord( 0,  1,  8));
            botHomes.push_back(Coord( 0,  1,  9));
            botHomes.push_back(Coord( 0,  1,  6));
            botHomes.push_back(Coord( 0,  1,  2));
            botHomes.push_back(Coord( 0,  1,  3));
            botHomes.push_back(Coord( 0,  1,  4));
            botHomes.push_back(Coord( 0,  1,  1));
        }
        botHomes.push_back(Coord(10,  0,  0));
        botHomes.push_back(Coord(15,  0,  0));
        botHomes.push_back(Coord(17,  0,  0));
        botHomes.push_back(Coord(18,  0,  0));
        botHomes.push_back(Coord(19,  0,  0));
        botHomes.push_back(Coord(16,  0,  0));
        botHomes.push_back(Coord(12,  0,  0));
        botHomes.push_back(Coord(13,  0,  0));
        botHomes.push_back(Coord(14,  0,  0));
        botHomes.push_back(Coord(11,  0,  0));
        botHomes.push_back(Coord( 5,  0,  0));
        botHomes.push_back(Coord( 7,  0,  0));
        botHomes.push_back(Coord( 8,  0,  0));
        botHomes.push_back(Coord( 9,  0,  0));
        botHomes.push_back(Coord( 6,  0,  0));
        botHomes.push_back(Coord( 2,  0,  0));
        botHomes.push_back(Coord( 3,  0,  0));
        botHomes.push_back(Coord( 4,  0,  0));
        botHomes.push_back(Coord( 1,  0,  0));
    }
    else
    {
        botHomes.push_back(Coord( 0,  0,  0));
    }

    vector<Coord> botCoords(botHomes);

    uint32_t res = targetModel.getRes();

    Model scopeModel;
    if (useScopeModel)
    {
        scopeModel.resize(res);
        for (uint32_t x = 0; x < res; x++)
        {
            for (uint32_t z = 0; z < res; z++)
            {
                scopeModel.set(x, 0, z, 1);
            }
        }
    }

    //printf("res = %" PRIu32 "\n", res);
    Model readyModel;
    uint32_t readyCount = 0;
    if (assemble && !erect)
    {
        readyModel.resize(res);
        for (uint32_t x = 1; x < res - 1; x++)
        {
            for (uint32_t z = 1; z < res - 1; z++)
            {
                if (targetModel.get(x, 0, z))
                {
                    readyModel.set(x, 0, z, 1);
                    readyCount++;
                }
            }
        }
        //printf("readyCount = %" PRIu32 "\n", readyCount);
    }

    Model buildModel(res);

    Model safeModel;
    if (assemble && !erect)
    {
        safeModel = readyModel;
    }
    uint32_t safeCount = readyCount;

    auto updateSafeModel = [&](Coord safeCoord)
    {
        if (safeModel.get(safeCoord))
        {
            safeCount--;
        }

        if (!readyModel.get(safeCoord))
        {
            safeModel.set(safeCoord, 0);
            return;
        }

        uint8_t box[3][3][3];
        for (uint32_t x = 0; x < 3; x++)
        {
            for (uint32_t y = 0; y < 3; y++)
            {
                for (uint32_t z = 0; z < 3; z++)
                {
                    Coord boxCoord = Coord(safeCoord.x + x - 1,
                                           safeCoord.y + y - 1,
                                           safeCoord.z + z - 1);
                    if (boxCoord.y < res)
                    {
                        box[x][y][z] = buildModel.get(boxCoord);
                    }
                    else
                    {
                        box[x][y][z] = 1;
                    }
                }
            }
        }
        box[1][1][1] = 1;

        uint8_t count = (
            box[0][1][1] +
            box[2][1][1] +
            box[1][0][1] +
            box[1][2][1] +
            box[1][1][0] +
            box[1][1][2]);
        if (count >= 5)
        {
            safeModel.set(safeCoord, 1);
            safeCount++;
            return;
        }

        Coord boxQ[27];
        if (box[0][1][1] == 0) boxQ[0] = Coord(0, 1, 1);
        else if (box[2][1][1] == 0) boxQ[0] = Coord(2, 1, 1);
        else if (box[1][0][1] == 0) boxQ[0] = Coord(1, 0, 1);
        else if (box[1][2][1] == 0) boxQ[0] = Coord(1, 2, 1);
        else if (box[1][1][0] == 0) boxQ[0] = Coord(1, 1, 0);
        else if (box[1][1][2] == 0) boxQ[0] = Coord(1, 1, 2);
        box[boxQ[0].x][boxQ[0].y][boxQ[0].z] = 2;
        uint32_t start = 0;
        uint32_t end = 1;
        while (start < end)
        {
            Coord coord = boxQ[start++];
            for (uint32_t adjDir = 0; adjDir < 6; adjDir++)
            {
                Coord adjCoord = coord + adjDeltas[adjDir];
                if (adjCoord.x >= 3 ||
                    adjCoord.y >= 3 ||
                    adjCoord.z >= 3)
                {
                    continue;
                }
                if (box[adjCoord.x][adjCoord.y][adjCoord.z] == 0)
                {
                    boxQ[end++] = adjCoord;
                    box[adjCoord.x][adjCoord.y][adjCoord.z] = 2;
                }
            }
        }

        if (box[0][1][1] == 0 ||
            box[2][1][1] == 0 ||
            box[1][0][1] == 0 ||
            box[1][2][1] == 0 ||
            box[1][1][0] == 0 ||
            box[1][1][2] == 0)
        {
            safeModel.set(safeCoord, 0);
        }
        else
        {
            safeModel.set(safeCoord, 1);
            safeCount++;
        }
    };

    uint32_t todoCount = 0;
    for (uint32_t x = 0; x < res; x++)
    {
        for (uint32_t y = 0; y < res; y++)
        {
            for (uint32_t z = 0; z < res; z++)
            {
                if (targetModel.get(x, y, z))
                {
                    todoCount++;
                }
            }
        }
    }

    //Coord coord = Coord();
    vector<Coord> dfsStack;
    Matrix<DfsEntry> dfsEntries(res);
    vector<QueueEntry> q; q.reserve(res * res * res);
    Model searchModel(res);
    vector<Command> tmpCommands;
    bool stuck = false;
    vector<Coord> volatileCoords;
    size_t volatileCoordsPos = 0;
    Model volatileModel(res);

    auto addVolatileCoords = [&](Coord coord, const Command& command)
    {
        if (command.m_action == Action::SMove)
        {
            Coord delta;
            uint32_t dist = 0;
            if (command.m_coord1.x != 0)
            {
                if (command.m_coord1.x <= 15)
                {
                    delta = Coord(1, 0, 0);
                    dist = command.m_coord1.x;
                }
                else
                {
                    delta = Coord(255, 0, 0);
                    dist = 256 - command.m_coord1.x;
                }
            }
            else if (command.m_coord1.y != 0)
            {
                if (command.m_coord1.y <= 15)
                {
                    delta = Coord(0, 1, 0);
                    dist = command.m_coord1.y;
                }
                else
                {
                    delta = Coord(0, 255, 0);
                    dist = 256 - command.m_coord1.y;
                }
            }
            else if (command.m_coord1.z != 0)
            {
                if (command.m_coord1.z <= 15)
                {
                    delta = Coord(0, 0, 1);
                    dist = command.m_coord1.z;
                }
                else
                {
                    delta = Coord(0, 0, 255);
                    dist = 256 - command.m_coord1.z;
                }
            }
            for (uint32_t i = 0; i < dist; i++)
            {
                coord += delta;
                volatileCoords.push_back(coord);
            }
        }
    };

    uint32_t numLayers = 0;
    if (disassemble)
    {
        for (auto& column : columnsD)
        {
            column.m_coordHigh.y++;
            numLayers += (column.m_coordHigh.y - column.m_coordLow.y);
        }

        printf("%8" PRIu32 "\x1b[8D", numLayers);
        fflush(stdout);
    }
    uint32_t layerPrintCount = 0;

    size_t numGroups = numBots / 4;
    vector<size_t> groupColumns(numGroups);
    size_t nextColumn = 0;
    for (size_t iGroup = 0; iGroup < numGroups; iGroup++)
    {
        if ((MAX_GROUPS > 0) && (iGroup >= MAX_GROUPS))
        {
            groupColumns[iGroup] = numColumnsD;
        }
        else
        {
            groupColumns[iGroup] = nextColumn++;
        }
    }
    vector<GroupState> groupStates(numGroups);
    vector<Coord> botTargets(numBots);

    vector<size_t> groupTasks(numGroups);
    size_t nextTask = 0;
    for (size_t iGroup = 0; iGroup < numGroups; iGroup++)
    {
        groupTasks[iGroup] = (size_t)-1;
    }
    vector<TaskState> taskStates(numGroups);

    if (disassemble)
    {
        for (auto& column : columnsD)
        {
            for (uint32_t x = column.m_coordLow.x; x <= column.m_coordHigh.x; x++)
            {
                uint32_t startY = column.m_coordLow.y;
#if 1
                if (startY > 2)
                {
                    startY = 2;
                }
#endif
                for (uint32_t y = startY; y < column.m_coordHigh.y; y++)
                {
                    for (uint32_t z = column.m_coordLow.z; z <= column.m_coordHigh.z; z++)
                    {
                        //buildModel.set(x, y, z, 1);
                        volatileModel.set(x, y, z, 1);
                    }
                }
            }
        }

        Coord bboxLow;
        Coord bboxHigh;
        findBBox(sourceModel, &bboxLow, &bboxHigh);
        for (uint32_t x = 0; x < res; x++)
        {
            for (uint32_t y = 2; y < res; y++)
            {
                for (uint32_t z = 0; z < res; z++)
                {
                    if (x + 1 < (uint32_t)bboxLow.x ||
                        x > (uint32_t)bboxHigh.x + 1 ||
                        //y + 1 < (uint32_t)bboxLow.y ||
                        y > (uint32_t)bboxHigh.y + 1 ||
                        z + 1 < (uint32_t)bboxLow.z ||
                        z > (uint32_t)bboxHigh.z + 1)
                    {
                        volatileModel.set(x, y, z, 1);
                    }
                }
            }
        }
    }

    uint32_t level = 0;
    vector<uint32_t> levelLeft(res);
    if (erect)
    {
        for (uint32_t x = 0; x < res; x++)
        {
            for (uint32_t y = 0; y < res; y++)
            {
                for (uint32_t z = 0; z < res; z++)
                {
                    if (targetModel.get(x, y, z))
                    {
                        levelLeft[y]++;
                    }
                }
            }
        }
    }
    bool levelStarted = false;

    vector<Task> tasks;
    bool needTasks = true;
    vector<uint8_t> tasksScratch;
    size_t tasksLeft = 0;

    //while (readyCount > 0 || coord != Coord(0, 0, 0))
    while (true)
    {
        bool goingHome = false;

        if (disassemble)
        {
            goingHome = true;

            for (size_t iGroup = 0; iGroup < numGroups; iGroup++)
            {
                size_t iColumn = groupColumns[iGroup];
                if (iColumn >= numColumnsD)
                {
                    botTargets[iGroup * 4 + 0] = botHomes[iGroup * 4 + 0];
                    botTargets[iGroup * 4 + 1] = botHomes[iGroup * 4 + 1];
                    botTargets[iGroup * 4 + 2] = botHomes[iGroup * 4 + 2];
                    botTargets[iGroup * 4 + 3] = botHomes[iGroup * 4 + 3];
                    groupStates[iGroup] = GroupState::Home;
                    continue;
                }

                goingHome = false;

                auto& column = columnsD[iColumn];

                botTargets[iGroup * 4 + 0] =
                    Coord(column.m_coordLow.x,
                          column.m_coordHigh.y,
                          column.m_coordLow.z);
                botTargets[iGroup * 4 + 1] =
                    Coord(column.m_coordHigh.x,
                          column.m_coordHigh.y,
                          column.m_coordLow.z);
                botTargets[iGroup * 4 + 2] =
                    Coord(column.m_coordLow.x,
                          column.m_coordHigh.y,
                          column.m_coordHigh.z);
                botTargets[iGroup * 4 + 3] =
                    Coord(column.m_coordHigh.x,
                          column.m_coordHigh.y,
                          column.m_coordHigh.z);

                if (botCoords[iGroup * 4 + 0] == botTargets[iGroup * 4 + 0] &&
                    botCoords[iGroup * 4 + 1] == botTargets[iGroup * 4 + 1] &&
                    botCoords[iGroup * 4 + 2] == botTargets[iGroup * 4 + 2] &&
                    botCoords[iGroup * 4 + 3] == botTargets[iGroup * 4 + 3])
                {
                    groupStates[iGroup] = GroupState::Void;
                }
                else
                {
                    groupStates[iGroup] = GroupState::Move;
                }
            }
        }
        uint32_t minMoveSteps = 9999999;
        uint32_t maxMoveSteps = 0;

        if (erect)
        {
            while (level < res - 1 && levelLeft[level] == 0 && tasksLeft == 0)
            {
                level++;
                needTasks = true;
            }
            if (level != res - 1)
            {
                levelStarted = true;
                if (levelLeft[level] != 0)
                {
                    for (size_t iBot = 0; iBot < numBots; iBot++)
                    {
                        if (botCoords[iBot].y != level + 1)
                        {
                            levelStarted = false;
                            break;
                        }
                    }
                }
            }

            if (level != res - 1 && levelStarted && needTasks)
            {
                tasks.clear();
                uint32_t numTasked = 0;
                uint32_t numUntasked = 0;
                if (USE_TASKS)
                {
                    findTasks(targetModel, tasksScratch, level, &tasks,
                              &numTasked, &numUntasked);
                }
                nextTask = 0;
                needTasks = false;
                tasksLeft = tasks.size();
                for (auto& task : tasks)
                {
                    uint32_t size =
                        (task.m_coordHigh.x - task.m_coordLow.x + 1) *
                        (task.m_coordHigh.z - task.m_coordLow.z + 1);
                    levelLeft[level] -= size;
                }

#if 0
                printf("\n");
                printf("%" PRIuZ " tasks, %" PRIu32 " tasked, %" PRIu32 " untasked\n", tasks.size(), numTasked, numUntasked);
                for (auto& task : tasks)
                {
                    printf("  %s - %s\n",
                           task.m_coordLow.toString().c_str(),
                           task.m_coordHigh.toString().c_str());
                }
#endif
                for (auto& task : tasks)
                {
                    for (uint32_t x = task.m_coordLow.x; x <= task.m_coordHigh.x; x++)
                    {
                        for (uint32_t z = task.m_coordLow.z; z <= task.m_coordHigh.z; z++)
                        {
                            buildModel.set(x, level, z, 1);
                        }
                    }
                }

                for (size_t iGroup = 0; iGroup < numGroups; iGroup++)
                {
                    if ((MAX_GROUPS > 0) && (iGroup >= MAX_GROUPS))
                    {
                        groupTasks[iGroup] = (size_t)-1;
                    }
                    else if (nextTask < tasks.size())
                    {
                        groupTasks[iGroup] = nextTask++;
                    }
                    else
                    {
                        groupTasks[iGroup] = (size_t)-1;
                    }
                }
            }

            for (size_t iGroup = 0; iGroup < numGroups; iGroup++)
            {
                size_t iTask = groupTasks[iGroup];
                if (iTask == (size_t)-1)
                {
                    //botTargets[iGroup * 4 + 0] = botHomes[iGroup * 4 + 0];
                    //botTargets[iGroup * 4 + 1] = botHomes[iGroup * 4 + 1];
                    //botTargets[iGroup * 4 + 2] = botHomes[iGroup * 4 + 2];
                    //botTargets[iGroup * 4 + 3] = botHomes[iGroup * 4 + 3];
                    taskStates[iGroup] = TaskState::None;
                    continue;
                }

                auto& task = tasks[iTask];

                botTargets[iGroup * 4 + 0] =
                    Coord(task.m_coordLow.x,
                          level,
                          task.m_coordLow.z);
                botTargets[iGroup * 4 + 1] =
                    Coord(task.m_coordHigh.x,
                          level,
                          task.m_coordLow.z);
                botTargets[iGroup * 4 + 2] =
                    Coord(task.m_coordLow.x,
                          level,
                          task.m_coordHigh.z);
                botTargets[iGroup * 4 + 3] =
                    Coord(task.m_coordHigh.x,
                          level,
                          task.m_coordHigh.z);

                bool ok = true;
                for (size_t iMember = 0; iMember < 4; iMember++)
                {
                    Coord botCoord = botCoords[iGroup * 4 + iMember];
                    Coord botTarget = botTargets[iGroup * 4 + iMember];
                    if (botCoord == botTarget + Coord(0, 1, 0) ||
                        botCoord == botTarget + Coord(-1, 1, 0) ||
                        botCoord == botTarget + Coord(1, 1, 0) ||
                        botCoord == botTarget + Coord(0, 1, -1) ||
                        botCoord == botTarget + Coord(0, 1, 1))
                    {
                    }
                    else
                    {
                        ok = false;
                        break;
                    }
                }
                if (ok)
                {
                    taskStates[iGroup] = TaskState::Fill;
                }
                else
                {
                    taskStates[iGroup] = TaskState::Move;
                }
            }
        }

        if (assemble && !erect && readyCount == 0)
        {
            goingHome = true;
        }

        if (assemble && erect && (level == res - 1))
        {
            goingHome = true;
        }

        if (goingHome)
        {
            bool allHome = true;
            for (size_t iBot = 0; iBot < numBots; iBot++)
            {
                if (botCoords[iBot] != botHomes[iBot])
                {
                    allHome = false;
                    break;
                }
            }
            if (allHome)
            {
                break;
            }
        }

#if 0
        if (assemble && !erect)
        {
            printf("readyCount = %" PRIu32 "\n", readyCount);
        }
        if (assemble && erect)
        {
            printf("level = %" PRIu32 ", tasksLeft = %" PRIuZ ", levelLeft[level] = %" PRIu32 ", levelStarted = %s\n", level, tasksLeft, levelLeft[level], levelStarted ? "true" : "false");
        }
        for (size_t iBot = 0; iBot < numBots; iBot++)
        {
            printf("%2" PRIuZ ": %s\n", iBot, botCoords[iBot].toString().c_str());
        }
#endif

        for (auto& volatileCoord : volatileCoords)
        {
            volatileModel.set(volatileCoord, 0);
        }
        volatileCoords.clear();
        volatileCoordsPos = 0;

        for (size_t iBot = 0; iBot < numBots; iBot++)
        {
            volatileCoords.push_back(botCoords[iBot]);
        }

        bool didSomething = false;

        for (size_t iBot = 0; iBot < numBots; iBot++)
        {
            bool useSafeCount =
                USE_SAFE_COUNT &&
                (safeCount > 0) &&
                ((SAFE_COUNT_FRACTION == 0) ||
                 ((iBot % SAFE_COUNT_FRACTION) != 0));

            // Find articulation points
#if 1
            if (assemble && !erect && !useSafeCount)
            {
                for (uint32_t x = 0; x < res; x++)
                {
                    for (uint32_t y = 0; y < res; y++)
                    {
                        for (uint32_t z = 0; z < res; z++)
                        {
                            auto& entry = dfsEntries(x, y, z);
                            entry.m_visited = 0;
                        }
                    }
                }

                {
                    dfsStack.push_back(Coord(0, 0, 0));
                    auto& entry = dfsEntries(Coord(0, 0, 0));
                    entry.m_articulation = 0;
                    entry.m_important = 0;
                    entry.m_parent = Coord(0, 0, 0);
                    entry.m_depth = 1;
                    entry.m_low = 1;
                }

                uint32_t dfsStartCount = 0;
                uint32_t dfsFinishCount = 0;
                while (!dfsStack.empty())
                {
                    Coord dfsCoord = dfsStack.back();
                    auto& entry = dfsEntries(dfsCoord);
                    if (entry.m_visited == 1)
                    {
                        dfsFinishCount++;
                        //printf("< %s\n", dfsCoord.toString().c_str());
                        entry.m_visited = 2;
                        if (entry.m_parent != Coord(0, 0, 0))
                        {
                            auto& parentEntry = dfsEntries(entry.m_parent);
                            //printf("parentEntry.m_low = %" PRIu32 "\n", parentEntry.m_low);
                            //printf("entry.m_depth = %" PRIu32 "\n", entry.m_depth);
                            if (entry.m_low >= parentEntry.m_depth /*&&
                                entry.m_important*/)
                            {
                                /*
                                  if (!parentEntry.m_articulation)
                                  {
                                  printf("%s: articulation = 1\n",
                                  entry.m_parent.toString().c_str());
                                  }
                                */
                                parentEntry.m_articulation =
                                    entry.m_important ? 2 : 1;
                            }
                            parentEntry.m_low = std::min(parentEntry.m_low, entry.m_low);
                            if (entry.m_important)
                            {
                                parentEntry.m_important = 1;
                            }
                        }
                    }
                    if (entry.m_visited)
                    {
                        dfsStack.pop_back();
                        continue;
                    }
                    dfsStartCount++;
                    //printf("> %s\n", dfsCoord.toString().c_str());
                    entry.m_visited = 1;
                    entry.m_articulation = 0;

                    for (uint32_t adjDir = 0; adjDir < 6; adjDir++)
                    {
                        Coord adjCoord = dfsCoord + adjDeltas[adjDir];
                        if (adjCoord.x >= res ||
                            adjCoord.y >= res ||
                            adjCoord.z >= res)
                        {
                            continue;
                        }
                        if (USE_SCOPE_MODEL && !scopeModel.get(adjCoord))
                        {
                            continue;
                        }
                        if (buildModel.get(adjCoord))
                        {
                            continue;
                        }
                        auto& adjEntry = dfsEntries(adjCoord);
                        if (!adjEntry.m_visited)
                        {
                            adjEntry.m_parent = dfsCoord;
                            adjEntry.m_depth = entry.m_depth + 1;
                            adjEntry.m_low = entry.m_depth + 1;
                            adjEntry.m_important = targetModel.get(adjCoord);
#if 1
                            for (size_t jBot = 0; jBot < numBots; jBot++)
                            {
                                if (jBot != iBot &&
                                    adjCoord == botCoords[jBot])
                                {
                                    adjEntry.m_important = 1;
                                }
                            }
#endif
                            dfsStack.push_back(adjCoord);
                        }
                        else if (adjCoord != entry.m_parent)
                        {
                            entry.m_low = std::min(entry.m_low, adjEntry.m_depth);
                        }
                    }
                }
                //printf("dfsStartCount = %" PRIu32 "\n", dfsStartCount);
                //printf("dfsFinishCount = %" PRIu32 "\n", dfsFinishCount);

                /*
                for (uint32_t x = 0; x < res; x++)
                {
                    for (uint32_t y = 0; y < res; y++)
                    {
                        for (uint32_t z = 0; z < res; z++)
                        {
                            if (targetModel.get(x, y, z) &&
                                !buildModel.get(x, y, z) &&
                                !dfsEntries(x, y, z).m_visited)
                            {
                                printf("Didn't visit %s\n",
                                       Coord(x, y, z).toString().c_str());
                            }
                        }
                    }
                }
                */
            }

#if 0
            {
                uint32_t fastSafeCount = 0;
                uint32_t slowSafeCount = 0;
                for (uint32_t x = 0; x < res; x++)
                {
                    for (uint32_t y = 0; y < res; y++)
                    {
                        for (uint32_t z = 0; z < res; z++)
                        {
                            if (!readyModel.get(x, y, z) &&
                                safeModel.get(x, y, z))
                            {
                                printf("safe but not ready?\n");
                            }
                            if (safeModel.get(x, y, z))
                            {
                                fastSafeCount++;
                            }
                            if (!readyModel.get(x, y, z))
                            {
                                continue;
                            }
                            if (!dfsEntries(x, y, z).m_articulation)
                            {
                                slowSafeCount++;
                            }
                        }
                    }
                }
                if (fastSafeCount != safeCount)
                {
                    printf("fastSafeCount != safeCount\n");
                }
                if (safeCount > slowSafeCount)
                {
                    printf("fast > slow\n");
                }
                if (safeCount > readyCount)
                {
                    printf("fast > ready\n");
                }
                if (slowSafeCount > readyCount)
                {
                    printf("slow > ready\n");
                }
                /*
                printf("Safe counts:  fast = %" PRIu32 ", slow = %" PRIu32 ", ready = %" PRIu32 "\n",
                       safeCount,
                       slowSafeCount,
                       readyCount);
                */
            }
#endif
#endif

            for (auto& entry : q)
            {
                searchModel.set(entry.m_coord.x,
                                entry.m_coord.y,
                                entry.m_coord.z,
                                0);
            }
            q.clear();

            while (volatileCoordsPos < volatileCoords.size())
            {
                volatileModel.set(volatileCoords[volatileCoordsPos++], 1);
            }

            size_t qpos = 0;

            {
                q.emplace_back();
                auto& entry = q.back();
                entry.m_coord = botCoords[iBot];
                entry.m_prev = 0;
                entry.m_command = Command(Action::Halt);
                searchModel.set(botCoords[iBot], 1);
            }

            size_t maxBots =
                (readyCount + (MAX_BOTS_FACTOR - 1)) / MAX_BOTS_FACTOR;

            Coord readyCoord;
            Coord readyDelta;
            size_t hpos = 0;
            bool foundHome = false;
            while (qpos < q.size())
            {
                /*
                if (qpos % 1024 == 0)
                {
                    printf(".");
                    fflush(stdout);
                }
                */
                uint32_t x = q[qpos].m_coord.x;
                uint32_t y = q[qpos].m_coord.y;
                uint32_t z = q[qpos].m_coord.z;
                //printf("x = %" PRIu32 ", y = %" PRIu32 ", z = %" PRIu32 "\n", x, y, z);

                if (q[qpos].m_coord == botHomes[iBot])
                {
                    hpos = qpos;
                    foundHome = true;
                }

                //if (readyCount == 0)
                //if (iBot >= readyCount)
                if (disassemble)
                {
                    if (q[qpos].m_coord == botTargets[iBot])
                    {
                        break;
                    }

                    /*
                    if (qpos > 0 &&
                        q[qpos].m_prev == 0 &&
                        botCoords[iBot].y + 15 <= botTargets[iBot].y &&
                        q[qpos].m_coord.x == botCoords[iBot].x &&
                        q[qpos].m_coord.y == botCoords[iBot].y + 15 &&
                        q[qpos].m_coord.z == botCoords[iBot].z)
                    {
                        break;
                    }
                    */
                }
                else if (erect && level == res - 1)
                {
                    if (q[qpos].m_coord == botHomes[iBot])
                    {
                        break;
                    }
                }
                else if (erect && !levelStarted)
                {
                    if (q[qpos].m_coord.y == level + 1)
                    {
                        break;
                    }
                }
                else if (erect && levelLeft[level] == 0 &&
                         taskStates[iBot / 4] == TaskState::None)
                {
                    if (level + 1 == res - 1)
                    {
                        if (q[qpos].m_coord == botHomes[iBot])
                        {
                            break;
                        }
                    }
                    else
                    {
                        if (q[qpos].m_coord.y == level + 2)
                        {
                            break;
                        }
                    }
                }
                else if (!erect && iBot >= maxBots)
                {
                    if (q[qpos].m_coord == botHomes[iBot])
                    {
                        break;
                    }
                }
                else
                {
                    bool foundReady = false;
                    for (uint32_t nearDir = 0; nearDir < 18; nearDir++)
                    {
                        Coord nearCoord = q[qpos].m_coord + nearDeltas[nearDir];
                        if (nearCoord.x >= res ||
                            nearCoord.y >= res ||
                            nearCoord.z >= res)
                        {
                            continue;
                        }
                        if (erect && nearCoord.y != level)
                        {
                            continue;
                        }

                        if (erect)
                        {
                            size_t iGroup = iBot / 4;
                            size_t iTask = groupTasks[iGroup];
                            if (iTask == (size_t)-1)
                            {
                                if (!targetModel.get(nearCoord) ||
                                    buildModel.get(nearCoord))
                                {
                                    continue;
                                }
                            }
                            else
                            {
                                if (nearCoord != botTargets[iBot])
                                {
                                    continue;
                                }
                            }
                        }
                        else
                        {
                            if (!readyModel.get(nearCoord) ||
                                !(/*qpos != 0 ||*/
                                  !volatileModel.get(nearCoord)))
                            {
                                continue;
                            }
                        }

                        if (erect)
                        {
                        }
                        else if (useSafeCount)
                        {
                            if (!safeModel.get(nearCoord))
                            {
                                continue;
                            }
                        }
                        else
                        {
                            auto& nearEntry = dfsEntries(nearCoord);
                            /*
                            if (!nearEntry.m_visited)
                            {
                                printf("!nearEntry.m_visited\n");
                            }
                            if (!dfsEntries(q[qpos].m_coord).m_visited)
                            {
                                printf("!dfsEntries(q[qpos].m_coord).m_visited\n");
                            }
                            */
                            if (!(!nearEntry.m_articulation ||
                                  (nearEntry.m_articulation == 1 &&
                                   nearEntry.m_depth >= dfsEntries(q[qpos].m_coord).m_depth)))
                            {
                                continue;
                            }
                        }

                        {
                            /*
                            printf("foundReady: %s\n", nearCoord.toString().c_str());
                            printf("q[qpos].m_coord: %s\n", q[qpos].m_coord.toString().c_str());
                            printf("m_articulation = %" PRIu8 "\n", nearEntry.m_articulation);
                            */
                            foundReady = true;
                            readyCoord = nearCoord;
                            readyDelta = nearDeltas[nearDir];
                            break;
                        }
                    }

                    if (foundReady)
                    {
                        break;
                    }
                }

                size_t shuffleStart = q.size();

                for (uint32_t dist = 1; dist <= 15; dist++)
                {
                    if (x < dist ||
                        //buildModel.get(x - dist, y, z) ||
                        volatileModel.get(x - dist, y, z))
                    {
                        break;
                    }
                    if (!searchModel.get(x - dist, y, z))
                    {
                        q.emplace_back();
                        auto& entry = q.back();
                        entry.m_coord = Coord(x - dist, y, z);
                        //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                        entry.m_prev = qpos;
                        entry.m_command = Command(Action::SMove,
                                                  Coord(-dist, 0, 0));
                        searchModel.set(x - dist, y, z, 1);
                    }
                }
                for (uint32_t dist = 1; dist <= 15; dist++)
                {
                    if (x + dist >= res ||
                        //buildModel.get(x + dist, y, z) ||
                        volatileModel.get(x + dist, y, z))
                    {
                        break;
                    }
                    if (erect)
                    {
                        if (level == res - 1)
                        {
                            if (x + dist > 1 && z > 1)
                            {
                                break;
                            }
                        }
                    }
                    if (!searchModel.get(x + dist, y, z))
                    {
                        q.emplace_back();
                        auto& entry = q.back();
                        entry.m_coord = Coord(x + dist, y, z);
                        //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                        entry.m_prev = qpos;
                        entry.m_command = Command(Action::SMove,
                                                  Coord(dist, 0, 0));
                        searchModel.set(x + dist, y, z, 1);
                    }
                }
                for (uint32_t dist = 1; dist <= 15; dist++)
                {
                    if (y < dist ||
                        //buildModel.get(x, y - dist, z) ||
                        volatileModel.get(x, y - dist, z))
                    {
                        break;
                    }
                    if (erect)
                    {
                        if (level == res - 1)
                        {
                            if (x > 1 && z > 1)
                            {
                                break;
                            }
                        }
                        else
                        {
                            if (y - dist < level + 1)
                            {
                                break;
                            }
                        }
                    }
                    if (!searchModel.get(x, y - dist, z))
                    {
                        q.emplace_back();
                        auto& entry = q.back();
                        entry.m_coord = Coord(x, y - dist, z);
                        //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                        entry.m_prev = qpos;
                        entry.m_command = Command(Action::SMove,
                                                  Coord(0, -dist, 0));
                        searchModel.set(x, y - dist, z, 1);
                    }
                }
                for (uint32_t dist = 1; dist <= 15; dist++)
                {
                    if (y + dist >= res ||
                        //buildModel.get(x, y + dist, z) ||
                        volatileModel.get(x, y + dist, z))
                    {
                        break;
                    }
                    if (erect)
                    {
                        if (level == res - 1)
                        {
                            if (x > 1 && z > 1)
                            {
                                break;
                            }
                        }
                        else if (levelLeft[level] == 0 &&
                                 taskStates[iBot / 4] == TaskState::None)
                        {
                            if (y + dist > level + 2)
                            {
                                break;
                            }
                        }
                        else
                        {
                            if (y + dist > level + 1)
                            {
                                break;
                            }
                        }
                    }
                    if (!searchModel.get(x, y + dist, z))
                    {
                        q.emplace_back();
                        auto& entry = q.back();
                        entry.m_coord = Coord(x, y + dist, z);
                        //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                        entry.m_prev = qpos;
                        entry.m_command = Command(Action::SMove,
                                                  Coord(0, dist, 0));
                        searchModel.set(x, y + dist, z, 1);
                    }
                }
                for (uint32_t dist = 1; dist <= 15; dist++)
                {
                    if (z < dist ||
                        //buildModel.get(x, y, z - dist) ||
                        volatileModel.get(x, y, z - dist))
                    {
                        break;
                    }
                    if (!searchModel.get(x, y, z - dist))
                    {
                        q.emplace_back();
                        auto& entry = q.back();
                        entry.m_coord = Coord(x, y, z - dist);
                        //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                        entry.m_prev = qpos;
                        entry.m_command = Command(Action::SMove,
                                                  Coord(0, 0, -dist));
                        searchModel.set(x, y, z - dist, 1);
                    }
                }
                for (uint32_t dist = 1; dist <= 15; dist++)
                {
                    if (z + dist >= res ||
                        //buildModel.get(x, y, z + dist) ||
                        volatileModel.get(x, y, z + dist))
                    {
                        break;
                    }
                    if (erect)
                    {
                        if (level == res - 1)
                        {
                            if (z + dist > 1 && x > 1)
                            {
                                break;
                            }
                        }
                    }
                    if (!searchModel.get(x, y, z + dist))
                    {
                        q.emplace_back();
                        auto& entry = q.back();
                        entry.m_coord = Coord(x, y, z + dist);
                        //printf("Possible coord: %s\n", entry.m_coord.toString().c_str());
                        entry.m_prev = qpos;
                        entry.m_command = Command(Action::SMove,
                                                  Coord(0, 0, dist));
                        searchModel.set(x, y, z + dist, 1);
                    }
                }

                if (shuffleStart < q.size())
                {
                    std::shuffle(q.begin() + shuffleStart, q.end(), gen);
                }

                qpos++;
            }

#if 1
            if (qpos == q.size() && botCoords[iBot] != botHomes[iBot])
            {
                if (!foundHome)
                {
#if 1
                    //printf("!foundHome (iBot = %" PRIuZ ")\n", iBot);
                    tmpCommands.push_back(Command(Action::Wait));
                    continue;
#endif

#if 0
                    printf("!foundHome\n");
                    stuck = true;
                    break;
#endif
                }
                while (q[hpos].m_prev != 0)
                {
                    hpos = q[hpos].m_prev;
                }
                auto& entry = q[hpos];
                tmpCommands.push_back(entry.m_command);
                addVolatileCoords(botCoords[iBot], entry.m_command);
                didSomething = true;
                /*
                printf("Coord %s -> %s\n",
                       botCoords[iBot].toString().c_str(),
                       q[hpos].m_coord.toString().c_str());
                */
                botCoords[iBot] = q[hpos].m_coord;
                continue;
            }
#endif

            if (qpos == q.size())
            {
#if 1
                tmpCommands.push_back(Command(Action::Wait));
                continue;
#endif

#if 0
                printf("stuck: qpos == q.size()\n");
                stuck = true;
                break;
#endif
            }

            if (useMulti)
            {
                if (qpos != 0)
                {
                    uint32_t curMoveSteps = 0;
                    while (q[qpos].m_prev != 0)
                    {
                        qpos = q[qpos].m_prev;
                        curMoveSteps++;
                    }
                    minMoveSteps = std::min(minMoveSteps, curMoveSteps);
                    maxMoveSteps = std::max(maxMoveSteps, curMoveSteps);
                    auto& entry = q[qpos];
                    tmpCommands.push_back(entry.m_command);
                    addVolatileCoords(botCoords[iBot], entry.m_command);
                    didSomething = true;
                    botCoords[iBot] = q[qpos].m_coord;
                    //printf("coord = %s\n", coord.toString().c_str());
                    continue;
                }
            }
            else
            {
                size_t tmpCommandsPos = tmpCommands.size();
                botCoords[iBot] = q[qpos].m_coord;
                while (qpos != 0)
                {
                    auto& entry = q[qpos];
                    tmpCommands.push_back(entry.m_command);
                    //addVolatileCoords(botCoords[iBot], entry.m_command);
                    didSomething = true;
                    qpos = entry.m_prev;
                }
                std::reverse(tmpCommands.begin() + tmpCommandsPos,
                             tmpCommands.end());
            }

            if (disassemble)
            {
                size_t iGroup = iBot / 4;
                size_t iMember = iBot % 4;

                if (groupStates[iGroup] != GroupState::Void)
                {
                    //printf("Bot %" PRIuZ ": not GroupState::Void\n", iBot);
                    tmpCommands.push_back(Command(Action::Wait));
                    continue;
                }
                //printf("Bot %" PRIuZ ": GroupState::Void\n", iBot);

                size_t iColumn = groupColumns[iGroup];
                auto& column = columnsD[iColumn];

                Coord nd = Coord(0, -1, 0);
                Coord fd;
                if (iMember == 0)
                {
                    fd = Coord(column.m_coordHigh.x - column.m_coordLow.x,
                               0,
                               column.m_coordHigh.z - column.m_coordLow.z);
                }
                else if (iMember == 1)
                {
                    fd = Coord(column.m_coordLow.x - column.m_coordHigh.x,
                               0,
                               column.m_coordHigh.z - column.m_coordLow.z);
                }
                else if (iMember == 2)
                {
                    fd = Coord(column.m_coordHigh.x - column.m_coordLow.x,
                               0,
                               column.m_coordLow.z - column.m_coordHigh.z);
                }
                else if (iMember == 3)
                {
                    fd = Coord(column.m_coordLow.x - column.m_coordHigh.x,
                               0,
                               column.m_coordLow.z - column.m_coordHigh.z);
                }

                tmpCommands.push_back(Command(Action::GVoid, nd, fd));
                didSomething = true;
                continue;
            }

            if (erect)
            {
                size_t iGroup = iBot / 4;
                size_t iMember = iBot % 4;

                if (taskStates[iGroup] == TaskState::None &&
                    levelLeft[level] == 0)
                {
                    tmpCommands.push_back(Command(Action::Wait));
                    continue;
                }

                if (taskStates[iGroup] == TaskState::Move)
                {
                    tmpCommands.push_back(Command(Action::Wait));
                    continue;
                }

                if (taskStates[iGroup] == TaskState::Fill)
                {
                    size_t iTask = groupTasks[iGroup];
                    auto& task = tasks[iTask];

                    Coord nd = botTargets[iBot] - botCoords[iBot];
                    Coord fd;
                    if (iMember == 0)
                    {
                        fd = Coord(task.m_coordHigh.x - task.m_coordLow.x,
                                   0,
                                   task.m_coordHigh.z - task.m_coordLow.z);
                    }
                    else if (iMember == 1)
                    {
                        fd = Coord(task.m_coordLow.x - task.m_coordHigh.x,
                                   0,
                                   task.m_coordHigh.z - task.m_coordLow.z);
                    }
                    else if (iMember == 2)
                    {
                        fd = Coord(task.m_coordHigh.x - task.m_coordLow.x,
                                   0,
                                   task.m_coordLow.z - task.m_coordHigh.z);
                    }
                    else if (iMember == 3)
                    {
                        fd = Coord(task.m_coordLow.x - task.m_coordHigh.x,
                                   0,
                                   task.m_coordLow.z - task.m_coordHigh.z);
                    }

                    tmpCommands.push_back(Command(Action::GFill, nd, fd));
                    didSomething = true;

                    continue;
                }
            }

            //if (readyCount == 0)
            //if (iBot >= readyCount)
            if (!erect && iBot >= maxBots)
            {
                // This should mean we are already home
                tmpCommands.push_back(Command(Action::Wait));
                continue;
            }

            if (erect && (level == res - 1 || !levelStarted))
            {
                tmpCommands.push_back(Command(Action::Wait));
                continue;
            }

            //printf("readyCoord = %s\n", readyCoord.toString().c_str());
            if (!erect)
            {
                readyModel.set(readyCoord, 0);
                readyCount--;
            }
            if (erect)
            {
                levelLeft[level]--;
            }
            buildModel.set(readyCoord, 1);
            volatileModel.set(readyCoord, 1);
            todoCount--;
            if (!erect)
            {
                for (uint32_t adjDir = 0; adjDir < 6; adjDir++)
                {
                    Coord adjCoord = readyCoord + adjDeltas[adjDir];
                    if (adjCoord.x >= res ||
                        adjCoord.y >= res ||
                        adjCoord.z >= res)
                    {
                        continue;
                    }
                    if (targetModel.get(adjCoord) &&
                        !buildModel.get(adjCoord) &&
                        !readyModel.get(adjCoord))
                    {
                        readyModel.set(adjCoord, 1);
                        readyCount++;
                    }
                }

                for (uint32_t x = 0; x < 3; x++)
                {
                    for (uint32_t y = 0; y < 3; y++)
                    {
                        for (uint32_t z = 0; z < 3; z++)
                        {
                            Coord safeCoord = readyCoord + Coord(x - 1, y - 1, z - 1);
                            if (safeCoord.x < res &&
                                safeCoord.y < res &&
                                safeCoord.z < res)
                            {
                                updateSafeModel(safeCoord);
                            }
                        }
                    }
                }
            }

            printf("%8" PRIu32 " %8" PRIu32 "\x1b[17D", todoCount, safeCount);
            fflush(stdout);

            tmpCommands.push_back(Command(Action::Fill, readyDelta));
            didSomething = true;

            if (useScopeModel)
            {
                for (uint32_t x = 0; x < 3; x++)
                {
                    for (uint32_t y = 0; y < 3; y++)
                    {
                        for (uint32_t z = 0; z < 3; z++)
                        {
                            Coord scopeCoord = readyCoord + Coord(x - 1, y - 1, z - 1);
                            if (scopeCoord.x < res &&
                                scopeCoord.y < res &&
                                scopeCoord.z < res)
                            {
                                scopeModel.set(scopeCoord, 1);
                            }
                        }
                    }
                }
            }
        }

        if (disassemble)
        {
            for (size_t iGroup = 0; iGroup < numGroups; iGroup++)
            {
                size_t iColumn = groupColumns[iGroup];
                if (iColumn >= numColumnsD)
                {
                    continue;
                }

                if (groupStates[iGroup] == GroupState::Void)
                {
                    auto& column = columnsD[iColumn];
                    column.m_coordHigh.y--;
                    for (uint32_t x = column.m_coordLow.x; x <= column.m_coordHigh.x; x++)
                    {
                        for (uint32_t z = column.m_coordLow.z; z <= column.m_coordHigh.z; z++)
                        {
                            //buildModel.set(x, column.m_coordHigh.y, z, 0);
                            volatileModel.set(x, column.m_coordHigh.y, z, 0);
                        }
                    }
                    if (column.m_coordHigh.y == column.m_coordLow.y)
                    {
                        groupColumns[iGroup] = nextColumn++;
                    }
                    numLayers--;
                }
            }

            printf("%8" PRIu32 " %8" PRIu32 " %8" PRIu32 " %c\x1b[28D",
                   numLayers,
                   maxMoveSteps,
                   minMoveSteps == 9999999 ? 0 : minMoveSteps,
                   "/-\\|"[layerPrintCount % 4]);
            fflush(stdout);
            layerPrintCount++;
        }

        if (erect)
        {
            for (size_t iGroup = 0; iGroup < numGroups; iGroup++)
            {
                if (taskStates[iGroup] == TaskState::Fill)
                {
                    size_t iTask = groupTasks[iGroup];
                    auto& task = tasks[iTask];

                    for (uint32_t x = task.m_coordLow.x; x <= task.m_coordHigh.x; x++)
                    {
                        for (uint32_t z = task.m_coordLow.z; z <= task.m_coordHigh.z; z++)
                        {
                            volatileModel.set(x, level, z, 1);
                        }
                    }

                    uint32_t size =
                        (task.m_coordHigh.x - task.m_coordLow.x + 1) *
                        (task.m_coordHigh.z - task.m_coordLow.z + 1);
                    //levelLeft[level] -= size;
                    todoCount -= size;
                    tasksLeft--;

                    printf("%8" PRIu32 " %8" PRIu32 "\x1b[17D", todoCount, safeCount);
                    fflush(stdout);

                    if (nextTask < tasks.size())
                    {
                        groupTasks[iGroup] = nextTask++;
                    }
                    else
                    {
                        groupTasks[iGroup] = (size_t)-1;
                    }
                }
            }
        }

        if (stuck || !didSomething)
        {
            break;
        }

#if 0
        printf("=== tmpCommands ===\n");
        for (auto& tmpCommand : tmpCommands)
        {
            printf("%s\n", tmpCommand.toString().c_str());
        }
        printf("===================\n");
#endif
        trace.insert(trace.end(), tmpCommands.begin(), tmpCommands.end());
        tmpCommands.clear();
    }

    bool allHome = true;
    for (size_t iBot = 0; iBot < numBots; iBot++)
    {
        if (botCoords[iBot] != botHomes[iBot])
        {
            allHome = false;
            break;
        }
    }

    if ((assemble && !erect && readyCount > 0) ||
        (assemble && erect && level != res - 1) ||
        !allHome)
    {
        printf("readyCount = %" PRIu32 ", allHome = %s\n",
               readyCount,
               allHome ? "true" : "false");
        /*
        for (size_t iBot = 0; iBot < numBots; iBot++)
        {
            printf("%2" PRIuZ ": %s\n", iBot, botCoords[iBot].toString().c_str());
        }
        */
        printf("Painted into a corner\n");
        //trace.clear();
        return;
    }

    if (useMulti)
    {
        // Fusion

        trace.push_back(Command(useHarmonics ? Action::Flip : Action::Wait));
        if (use40)
        {
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            trace.push_back(Command(Action::Wait));
        }
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        trace.push_back(Command(Action::Wait));

        //  1         ( 0,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        if (use40)
        {
            //  2         ( 0,  1,  0)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            //  3         ( 0,  1, 10)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            //  4         ( 0,  1, 15)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            //  5         ( 0,  1, 17)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            //  6   7- 7  ( 0,  1, 18)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            //  8         ( 0,  1, 16)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            //  9         ( 0,  1, 12)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            // 10  11-11  ( 0,  1, 13)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            // 12         ( 0,  1, 11)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            // 13         ( 0,  1,  5)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            // 14         ( 0,  1,  7)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            // 15  16-16  ( 0,  1,  8)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            // 17         ( 0,  1,  6)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            // 18         ( 0,  1,  2)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            // 19  20-20  ( 0,  1,  3)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            // 21         ( 0,  1,  1)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
        }
        // 22         (10,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 23         (15,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 24         (17,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 25  26-26  (18,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 27         (16,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 28         (12,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 29  30-30  (13,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 31         (11,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 32         ( 5,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 33         ( 7,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 34  35-35  ( 8,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 36         ( 6,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 37         ( 2,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 38  39-39  ( 3,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 40         ( 1,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));

        trace.push_back(Command(Action::Wait));
        if (use40)
        {
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::SMove, Coord(0, 0, -1)));
            trace.push_back(Command(Action::SMove, Coord(0, 0, -1)));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::SMove, Coord(0, 0, -1)));
            trace.push_back(Command(Action::SMove, Coord(0, 0, -1)));
        }
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::SMove, Coord(-1, 0, 0)));
        trace.push_back(Command(Action::SMove, Coord(-1, 0, 0)));
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::SMove, Coord(-1, 0, 0)));
        trace.push_back(Command(Action::SMove, Coord(-1, 0, 0)));

        //  1  40-40  ( 0,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        if (use40)
        {
            //  2  21-21  ( 0,  1,  0)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            //  3  12-12  ( 0,  1, 10)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            //  4   8- 8  ( 0,  1, 15)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            //  5   6- 7  ( 0,  1, 16)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            //  9  10-11  ( 0,  1, 11)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            // 13  17-17  ( 0,  1,  5)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            // 14  15-16  ( 0,  1,  6)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            // 18  19-20  ( 0,  1,  1)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
        }
        // 22  31-31  (10,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 23  27-27  (15,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 24  25-26  (16,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 28  29-30  (11,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 32  36-36  ( 5,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 33  34-35  ( 6,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 37  38-39  ( 1,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));

        trace.push_back(Command(Action::Wait));
        if (use40)
        {
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::SMove, Coord(0, 0, -4)));
            trace.push_back(Command(Action::SMove, Coord(0, 0, -4)));
        }
        trace.push_back(Command(Action::Wait));
        trace.push_back(Command(Action::SMove, Coord(-4, 0, 0)));
        trace.push_back(Command(Action::SMove, Coord(-4, 0, 0)));

        //  1  37-40  ( 0,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        if (use40)
        {
            //  2  18-21  ( 0,  1,  0)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            //  3   9-12  ( 0,  1, 10)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            //  4   5- 8  ( 0,  1, 11)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
            // 13  14-17  ( 0,  1,  1)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
        }
        // 22  28-31  (10,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        // 23  24-27  (11,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));
        // 32  33-36  ( 1,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));

        trace.push_back(Command(Action::Wait));
        if (use40)
        {
            trace.push_back(Command(Action::Wait));
            trace.push_back(Command(Action::SMove, Coord(0, 0, -9)));
        }
        trace.push_back(Command(Action::SMove, Coord(-9, 0, 0)));

        //  1  32-40  ( 0,  0,  0)
        trace.push_back(Command(Action::FusionP, Coord(1, 0, 0)));
        if (use40)
        {
            //  2  13-21  ( 0,  1,  0)
            trace.push_back(Command(Action::FusionP, Coord(0, 0, 1)));
            //  3   4-12  ( 0,  1,  1)
            trace.push_back(Command(Action::FusionS, Coord(0, 0, -1)));
        }
        // 22  23-31  ( 1,  0,  0)
        trace.push_back(Command(Action::FusionS, Coord(-1, 0, 0)));

        if (use40)
        {
            //  1  22-40  ( 0,  0,  0)
            trace.push_back(Command(Action::FusionP, Coord(0, 1, 0)));
            //  2   3-21  ( 0,  1,  0)
            trace.push_back(Command(Action::FusionS, Coord(0, -1, 0)));
        }

        //  1   2-40  ( 0,  0,  0)
    }

    trace.push_back(Command(Action::Halt));
}
