#include "Common.hpp"
#include "StringUtils.hpp"
#include "Model.hpp"
#include "Command.hpp"
#include "Trace.hpp"
#include "Sim.hpp"
#include "Check.hpp"
#include "Solve.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage: score <batch> [<start> [<end>]]\n");
    printf("Batches:\n");
    printf("  la    Lightning assemble  (range: 1-186)\n");
    printf("  fa    Full assemble       (range: 1-186)\n");
    printf("  fd    Full disassemble    (range: 1-186)\n");
    printf("  fr    Full reassemble     (range: 1-115)\n");
}

int main(int argc, char *argv[])
{
    if (argc < 2 || argc > 4)
    {
        usage();
        return 1;
    }

    string batch = argv[1];
    uint32_t start = 1;
    uint32_t end = 186;
    bool full = true;
    if (batch == "la")
    {
        full = false;
    }
    else if (batch == "fa")
    {
    }
    else if (batch == "fd")
    {
    }
    else if (batch == "fr")
    {
        end = 115;
    }
    else
    {
        usage();
        return 1;
    }

    if (argc >= 3)
    {
        uint32_t tmpStart = (uint32_t)atoi(argv[2]);
        if (tmpStart < start || tmpStart > end)
        {
            usage();
            return 1;
        }
        start = tmpStart;
    }

    if (argc >= 4)
    {
        uint32_t tmpEnd = (uint32_t)atoi(argv[3]);
        if (tmpEnd < start || tmpEnd > end)
        {
            usage();
            return 1;
        }
        end = tmpEnd;
    }

    string msg;

    uint64_t totalScoreLoss = 0;
    for (uint32_t i = start; i <= end; i++)
    {
        printf("\n");

        Model sourceModel;
        Model targetModel;
        string defaultTraceName;
        string bestTraceName;

        if (batch == "la")
        {
            string targetModelName = strprintf("../../../data/problemsL/LA%03" PRIu32 "_tgt.mdl", i);
            if (!targetModel.load(targetModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            sourceModel.resize(targetModel.getRes());

            defaultTraceName = strprintf("../../../data/dfltTracesL/LA%03" PRIu32 ".nbt", i);
            bestTraceName = strprintf("../../../data/bestL/LA%03" PRIu32 ".nbt", i);
        }

        if (batch == "fa")
        {
            string targetModelName = strprintf("../../../data/problemsF/FA%03" PRIu32 "_tgt.mdl", i);
            if (!targetModel.load(targetModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            sourceModel.resize(targetModel.getRes());

            defaultTraceName = strprintf("../../../data/dfltTracesF/FA%03" PRIu32 ".nbt", i);
            bestTraceName = strprintf("../../../data/bestF/FA%03" PRIu32 ".nbt", i);
        }

        if (batch == "fd")
        {
            string sourceModelName = strprintf("../../../data/problemsF/FD%03" PRIu32 "_src.mdl", i);
            if (!sourceModel.load(sourceModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            targetModel.resize(sourceModel.getRes());

            defaultTraceName = strprintf("../../../data/dfltTracesF/FD%03" PRIu32 ".nbt", i);
            bestTraceName = strprintf("../../../data/bestF/FD%03" PRIu32 ".nbt", i);
        }

        if (batch == "fr")
        {
            string sourceModelName = strprintf("../../../data/problemsF/FR%03" PRIu32 "_src.mdl", i);
            if (!sourceModel.load(sourceModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            string targetModelName = strprintf("../../../data/problemsF/FR%03" PRIu32 "_tgt.mdl", i);
            if (!targetModel.load(targetModelName, &msg))
            {
                printf("%03" PRIu32": %s\n", i, msg.c_str());
                continue;
            }

            if (sourceModel.getRes() != targetModel.getRes())
            {
                printf("%03" PRIu32": Resolution mismatch\n", i);
                continue;
            }

            defaultTraceName = strprintf("../../../data/dfltTracesF/FR%03" PRIu32 ".nbt", i);
            bestTraceName = strprintf("../../../data/bestF/FR%03" PRIu32 ".nbt", i);
        }


        // Best score

        uint64_t maxScore = 0;
        {
            uint32_t res = targetModel.getRes();
            while (res > 0)
            {
                maxScore++;
                res >>= 1;
            }
            maxScore--;
            maxScore *= 1000;
        }


        // Default trace

        vector<Command> defaultTrace;
        if (!loadTrace(full, defaultTraceName, &defaultTrace, &msg))
        {
            printf("%03" PRIu32": %s\n", i, msg.c_str());
            continue;
        }

#if 0
        for (auto& command : defaultTrace)
        {
            printf("%s\n", command.toString().c_str());
        }
#endif

        printf("%03" PRIu32": Default trace energy:", i);
        fflush(stdout);

        uint64_t defaultEnergy = 0;
#if 1
        if (!checkTrace(full, sourceModel, targetModel, defaultTrace, &defaultEnergy, &msg))
        {
            printf("\n");
            printf("%03" PRIu32": %s\n", i, msg.c_str());
            continue;
        }
#endif
        printf("  %16" PRIu64 "  %4" PRIu64 "\n", defaultEnergy, maxScore);


        // Best trace

        vector<Command> bestTrace;
        if (!loadTrace(full, bestTraceName, &bestTrace, &msg))
        {
            printf("%03" PRIu32": %s\n", i, msg.c_str());
            continue;
        }

        printf("%03" PRIu32": Best trace energy:", i);
        fflush(stdout);

        uint64_t bestEnergy = 0;
#if 1
        if (!checkTrace(full, sourceModel, targetModel, bestTrace, &bestEnergy, &msg))
        {
            printf("\n");
            printf("%03" PRIu32": %s\n", i, msg.c_str());
            continue;
        }
#endif
        //uint64_t bestScore = maxScore * (defaultEnergy - bestEnergy) / defaultEnergy;
        uint64_t bestScore = (uint64_t)((__uint128_t)maxScore * (__uint128_t)(defaultEnergy - bestEnergy) / (__uint128_t)defaultEnergy);
        printf("     %16" PRIu64 "  %4" PRIu64 "\n", bestEnergy, bestScore);
        uint64_t scoreLoss = maxScore - bestScore;
        printf("%03" PRIu32": Score loss:                              %4" PRIu64 "\n", i, scoreLoss);
        totalScoreLoss += scoreLoss;
    }

    printf("\n");
    printf("%3s: Total score loss:                     %7" PRIu64 "\n", batch.c_str(), totalScoreLoss);
    printf("\n");

    return 0;
}
